
/*===================================================================
 *
 *    Antenna Positions Optimization
 *
 *   Date: sep 2000
 *   Author: F. Boone
 *   Purpose: Optimize the antenna positions of an interferometer for
 *            a given model distribution of visibilities.
 *
 *====================================================================*/

#include "APO.h"


/*======================================================================
 *                            FUNCTIONS 
 *======================================================================*/

void checkimplemented(APO *vapo, int napo, 
		      Instrume *pinstr[], Obs *pobs[], Site *psite[], Grid *pgrid[],
		      Transf *ptransf[],
		      int *sameinstr, int *sameobs, int *samesite, int *samegrid){

  int iapo;

  *sameinstr=1;
  *sameobs=1;
  *samesite=1;
  *samegrid=1;
  for(iapo=0;iapo<napo;iapo++) {
    pinstr[iapo]=vapo[iapo].getpinstr();
    pobs[iapo]=vapo[iapo].getpobs();
    ptransf[iapo]=vapo[iapo].gettransf();
    psite[iapo]=vapo[iapo].getpsite();
    pgrid[iapo]=vapo[iapo].getpgrid();
  }
  if(napo>0){
    iapo=1;
    while(iapo<napo && *sameinstr==1){
      if(pinstr[iapo]!=pinstr[0]) *sameinstr=0;
      if(pobs[iapo]!=pobs[0]) *sameobs=0;
      if(psite[iapo]!=psite[0]) *samesite=0;
      if(pgrid[iapo]!=pgrid[0]) *samegrid=0;
      iapo++;
    }
  }
  if(*sameinstr==0) {
    printf("Several instruments.\n");
    printf("  code not yet developed for this situation.\n");
    exit(1);
  }
  if(*sameobs==0) {
    printf("Several observation situations.\n");
    printf("  code not yet developed for this situation.\n");
    exit(1);
  }
   if(*samesite==0) {
    printf("Several sites.\n");
    printf("  code not yet developed for this situation.\n");
    exit(1);
  }   
   if(*samegrid==0) {
    printf("Several grids.\n");
  }  
}

void optimize(APO *vapo,int napo, float fa, float dptacc){
  Instrume *pinstr[napo];
  Obs *pobs[napo];
  Site *psite[napo];
  Grid *pgrid[napo];
  Transf *ptransf[napo];
  float **dpt, *dpt0;
  int nmoved,k,l,natot,nok,iapo,indapo;
  int sameinstr, sameobs, samesite, samegrid;
  int couldmove, ind, iant;
  float tmpfu, tmpfv, tmpfu0, tmpfv0, dpt1;
  char pause;
  printf("startoptimize\n");

  checkimplemented(vapo, napo, pinstr, pobs, psite, pgrid, ptransf,
		      &sameinstr, &sameobs, &samesite, &samegrid);
  natot=pinstr[0]->getnatot();
  if(natot>=NATOTMAX){
    printf("ERROR in APO::optimize : too many antennas!!!\n");
    exit(1);
  }
  dpt=matrix(natot,3);
  dpt0=dpt[0];
  nmoved=0;
  indapo=0;

  nok=1;
  while(nmoved < natot && nok!=0) {
    nok=0;
    for(iant=0;iant<natot;iant++){
      tmpfu=0.0; tmpfv=0.0;
      for(iapo=0;iapo<napo;iapo++){
	vapo[iapo].getforce(&tmpfu0, &tmpfv0, iant,fa);
	tmpfu+=tmpfu0; tmpfv+=tmpfv0;
      }
      tmpfu/=napo; tmpfv/=napo;
      dpt1=tmpfu*tmpfu+tmpfv*tmpfv;
      if(dpt1>0.0){
	dpt[nok][0]=tmpfu; 
	dpt[nok][1]=tmpfv;
	dpt[nok][2]=iant;
      	nok++;
      }
    }
    //printf("nok=%d\n",nok);
    qsort(dpt,nok,sizeof(float *),compvec);
    couldmove=0; k=nok-1; l=0;
    while(!couldmove && k>=l && nmoved<natot){ 
      tmpfu=dpt[k][0]; tmpfv=dpt[k][1]; ind=(int) dpt[k][2];
      couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, tmpfu, tmpfv, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
      if(!couldmove && k==0) {
	dpt1=sqrt(tmpfu*tmpfu+tmpfv*tmpfv);
	tmpfu=-tmpfu*2.1*dptacc/dpt1; 
	tmpfv=-tmpfv*2.1*dptacc/dpt1; 
	couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, tmpfu, tmpfv, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
	if(!couldmove) {
	  couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, 0., 0.,dptacc,
		     sameinstr, sameobs, samesite, samegrid);
	  nmoved++; k+=2; l++;
	}
      }
      k--;
      if(k==nok) nmoved=natot;
    }
    printf("%d trials \n",nok-k-2);				
    if(couldmove) nmoved++; 
  }
  pinstr[indapo]->resetmoved();  
  dpt[0]=dpt0;
  delete_matrix(dpt);
  dpt=NULL;
}

void optimize_shad(APO *vapo,int napo, float fa, float dptacc, float gsh){
  Instrume *pinstr[napo];
  Obs *pobs[napo];
  Site *psite[napo];
  Grid *pgrid[napo];
  Transf *ptransf[napo];
  float **dpt, *dpt0;
  int nmoved,k,l,natot,nok,iapo,indapo;
  int sameinstr, sameobs, samesite, samegrid;
  int couldmove, ind, iant;
  float tmpfu, tmpfv, tmpfu0, tmpfv0, dpt1;
  char pause;
  printf("startoptimize\n");

  checkimplemented(vapo, napo, pinstr, pobs, psite, pgrid, ptransf,
		      &sameinstr, &sameobs, &samesite, &samegrid);

  natot=pinstr[0]->getnatot();
  if(natot>=NATOTMAX){
    printf("ERROR in APO::optimize : too many antennas!!!\n");
    exit(1);
  }
  dpt=matrix(natot,3);
  dpt0=dpt[0];
  nmoved=0;
  indapo=0;
  nok=1;
  while(nmoved < natot && nok!=0) {
    nok=0;
    for(iant=0;iant<natot;iant++){
      tmpfu=0.0; tmpfv=0.0;
      for(iapo=0;iapo<napo;iapo++){
	vapo[iapo].getforce_shad(&tmpfu0, &tmpfv0, iant,fa, gsh);
	tmpfu+=tmpfu0; tmpfv+=tmpfv0;
      }
      tmpfu/=napo; tmpfv/=napo;
      dpt1=tmpfu*tmpfu+tmpfv*tmpfv;
      if(dpt1>0.0){
	dpt[nok][0]=tmpfu; 
	dpt[nok][1]=tmpfv;
	dpt[nok][2]=iant;
      	nok++;
      }
    }
    //printf("nok=%d\n",nok);
    qsort(dpt,nok,sizeof(float *),compvec);
    couldmove=0; k=nok-1; l=0;
    while(!couldmove && k>=l && nmoved<natot){ 
      tmpfu=dpt[k][0]; tmpfv=dpt[k][1]; ind=(int) dpt[k][2];
      couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, tmpfu, tmpfv, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
      if(!couldmove && k==0) {
	dpt1=sqrt(tmpfu*tmpfu+tmpfv*tmpfv);
	tmpfu=-tmpfu*2.1*dptacc/dpt1; 
	tmpfv=-tmpfv*2.1*dptacc/dpt1; 
	couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, tmpfu, tmpfv, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
	if(!couldmove) {
	  couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, 0., 0.,dptacc,
		     sameinstr, sameobs, samesite, samegrid);
	  nmoved++; k+=2; l++;
	}
      }
      k--;
      if(k==nok) nmoved=natot;
    }
    printf("%d trials \n",nok-k-2);				
    if(couldmove) nmoved++; 
  }
  pinstr[indapo]->resetmoved();  
  dpt[0]=dpt0;
  delete_matrix(dpt);
  dpt=NULL;
}

void optimize(APO *vapo,int napo, float fa, float dptacc, int conf){
  Instrume *pinstr[napo];
  Obs *pobs[napo];
  Site *psite[napo];
  Grid *pgrid[napo];
  Transf *ptransf[napo];
  float **dpt, *dpt0;
  int i,nmoved,k,l,na, nsh, nok,iapo,indapo;
  int sameinstr, sameobs, samesite, samegrid;
  int couldmove, ind, iant;
  float tmpfu, tmpfv, tmpfu0, tmpfv0, dpt1;
  char pause;
  printf("startoptimize\n");

  checkimplemented(vapo, napo, pinstr, pobs, psite, pgrid, ptransf,
		      &sameinstr, &sameobs, &samesite, &samegrid);
 
  indapo=0;
  na=pinstr[indapo]->getna();
  nsh=pinstr[indapo]->getnsh();
  dpt=matrix(na,3);
  dpt0=dpt[0];
  nmoved=0;
  nok=1;
  while(nmoved < na && nok!=0) {
    nok=0;
    for(i=0;i<na;i++){
      iant=conf*(na-nsh)+i;
      tmpfu=0.0; tmpfv=0.0;
      for(iapo=0;iapo<napo;iapo++){
	vapo[iapo].getforce(&tmpfu0, &tmpfv0, iant,fa);
	tmpfu+=tmpfu0; tmpfv+=tmpfv0;
      }
      tmpfu/=napo; tmpfv/=napo;
      dpt1=tmpfu*tmpfu+tmpfv*tmpfv;
      if(dpt1>0.0){
	dpt[nok][0]=tmpfu; 
	dpt[nok][1]=tmpfv;
	dpt[nok][2]=iant;
      	nok++;
      }
    }
    //printf("nok=%d\n",nok);
    qsort(dpt,nok,sizeof(float *),compvec);
    couldmove=0; k=nok-1; l=0;
    while(!couldmove && k>=l && nmoved<na){ 
      tmpfu=dpt[k][0]; tmpfv=dpt[k][1]; ind=(int) dpt[k][2];
      couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, tmpfu, tmpfv, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
      if(!couldmove && k==0) {
	dpt1=sqrt(tmpfu*tmpfu+tmpfv*tmpfv);
	tmpfu=-tmpfu*2.1*dptacc/dpt1; 
	tmpfv=-tmpfv*2.1*dptacc/dpt1; 
	couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, tmpfu, tmpfv, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
	if(!couldmove) {
	  couldmove=move(vapo, pinstr, pobs, ptransf,psite,
		     napo, indapo, ind, 0., 0.,dptacc,
		     sameinstr, sameobs, samesite, samegrid);
	  nmoved++; k+=2; l++;
	}
      }
      k--;
      if(k==nok) nmoved=na;
    }
    printf("%d trials \n",nok-k-2);				
    if(couldmove) nmoved++; 
  }
  pinstr[indapo]->resetmoved();  
  dpt[0]=dpt0;
  delete_matrix(dpt);
  dpt=NULL;
}

void shake(APO *vapo,int napo, float dptacc){
  Instrume *pinstr[napo];
  Obs *pobs[napo];
  Site *psite[napo];
  Grid *pgrid[napo];
  Transf *ptransf[napo];
  int i,iant,natot,iapo,indapo;
  int sameinstr, sameobs, samesite, samegrid;
  int couldmove;
  float a,t,randmax,dx,dy;

  printf("start shake\n");
  indapo=0;
  randmax=(float) RAND_MAX;
  srand(time(NULL));

  checkimplemented(vapo, napo, pinstr, pobs, psite, pgrid, ptransf,
		      &sameinstr, &sameobs, &samesite, &samegrid);
 
  natot=pinstr[0]->getnatot();
  if(natot>=NATOTMAX){
    printf("ERROR in APO::optimize : too many antennas!!!\n");
    exit(1);
  }


  for(iant=0;iant < natot;iant++) {
    if((pinstr[0]->getfixed(iant))==1) continue ;
    couldmove=0;
    i=0;
    while(!couldmove && i<20){
      a=(1+rand()*1.6/randmax)*dptacc;
      t=(-180+rand()*360./randmax)*PI/180.;
      dx=a*cos(t); dy=a*sin(t);
      couldmove=move(vapo, pinstr, pobs, ptransf, psite,
		     napo, indapo, iant, dx, dy, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
      if(couldmove) printf("\n");
      i++;
    }
  }
  pinstr[indapo]->resetmoved();  

}

void correct(APO *vapo,int napo, float dptacc){
  Instrume *pinstr[napo];
  Obs *pobs[napo];
  Site *psite[napo];
  Grid *pgrid[napo];
  Transf *ptransf[napo];
  int i,iant=0,iant0=0,natot,iapo,indapo;
  int sameinstr, sameobs, samesite, samegrid;
  int couldmove;
  long ncor=0l, ntrials=0l;
  float a,t,randmax,dx,dy,dx2,dy2;
  couldmove=0;

  printf("start correct\n");
  printf("dptacc=%f\n",dptacc);

  checkimplemented(vapo, napo, pinstr, pobs, psite, pgrid, ptransf,
		      &sameinstr, &sameobs, &samesite, &samegrid);
  printf("ok\n");

  indapo=0;
  randmax=(float) RAND_MAX;
  srand(time(NULL));

  natot=pinstr[0]->getnatot();
  if(natot>=NATOTMAX){
    printf("ERROR in APO::optimize : too many antennas!!!\n");
    exit(1);
  }

  while((iant!=-1 || iant0!=0) && ntrials<20*natot){
    //    printf("ncor=%ld iant=%d\n",ncor,iant);

    if(iant!=-1 && !couldmove) iant0=(iant+1)%natot;
    if(iant==-1 && iant0!=0) iant0=0;

    //printf("natot=%d iant0=%d\n",natot,iant0);
    pinstr[indapo]->getcorrection(ptransf[indapo], psite[indapo], 
				  iant0, &iant, &dx, &dy);
    //printf("\n ncor=%ld iant=%d dx=%f dy=%f\n",ncor,iant,dx,dy);
    if(iant!=-1){
      couldmove=move(vapo, pinstr, pobs, ptransf, psite,
		     napo, indapo, iant, dx, dy, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
      i=0;
      while(!couldmove && i<20){
	a=1+rand()*1.2/randmax;
	t=(-45+rand()*90./randmax)*PI/180.;
	dx2=a*dx*cos(t)+a*dy*sin(t);
	dy2=-a*dx*sin(t)+a*dy*cos(t);
	couldmove=move(vapo, pinstr, pobs, ptransf, psite,
		       napo, indapo, iant, dx2, dy2, dptacc,
		       sameinstr, sameobs, samesite, samegrid);
	i++;
      }
      if(couldmove) {
	//printf("moved\n");
	iant0=0;
	printf("\n");
	ncor++;
      }
    }
    //else if(ntrials==0) couldmove=1;
    ntrials++;
  }
  //printf("couldmove=%d iant=%d ntrials=%ld\n",couldmove,iant,ntrials);
  if((iant!=-1 || iant0!=0) && ntrials>=20*natot){
    printf("COULD NOT CORRECT CONFIGURATION!!!!!!\n");
    exit(1);
    //return;
  }
  pinstr[indapo]->resetmoved();  
  printf("%ld corrections were made.\n",ncor);

}

void correct(APO *vapo,int napo, float dptacc, int conf){
  Instrume *pinstr[napo];
  Obs *pobs[napo];
  Site *psite[napo];
  Grid *pgrid[napo];
  Transf *ptransf[napo];
  int i,iant=0,iant0=0,na,iapo,indapo;
  int sameinstr, sameobs, samesite, samegrid;
  int couldmove;
  long ncor=0l, ntrials=0l;
  float a,t,randmax,dx,dy,dx2,dy2;
  couldmove=0;

  printf("start correct\n");
  indapo=0;
  randmax=(float) RAND_MAX;
  srand(time(NULL));

  checkimplemented(vapo, napo, pinstr, pobs, psite, pgrid, ptransf,
		      &sameinstr, &sameobs, &samesite, &samegrid);

  na=pinstr[0]->getna();

  while((iant!=-1 || iant0!=0) && ntrials<20*na){
    //    printf("ncor=%ld iant=%d\n",ncor,iant);
    if(iant!=-1 && !couldmove) iant0=(iant+1)%na;
    if(iant==-1 && iant0!=0) iant0=0;

    //printf("natot=%d iant0=%d\n",natot,iant0);
    pinstr[indapo]->getcorrection(ptransf[indapo], psite[indapo], 
				  iant0, conf, &iant, &dx, &dy);
    //  printf("\n ncor=%ld iant=%d dx=%f dy=%f\n",ncor,iant,dx,dy);
    if(iant!=-1){
      couldmove=move(vapo, pinstr, pobs, ptransf, psite,
		     napo, indapo, iant, dx, dy, dptacc,
		     sameinstr, sameobs, samesite, samegrid);
      i=0;
      while(!couldmove && i<20){
	a=1+rand()*1.2/randmax;
	t=(-45+rand()*90./randmax)*PI/180.;
	dx2=a*dx*cos(t)+a*dy*sin(t);
	dy2=-a*dx*sin(t)+a*dy*cos(t);
	couldmove=move(vapo, pinstr, pobs, ptransf, psite,
		       napo, indapo, iant, dx2, dy2, dptacc,
		       sameinstr, sameobs, samesite, samegrid);
	i++;
      }
      if(couldmove) {
	//printf("moved\n");
	iant0=0;
	printf("\n");
	ncor++;
      }
    }
    //else if(ntrials==0) couldmove=1;
    ntrials++;
  }
  //printf("couldmove=%d iant=%d ntrials=%ld\n",couldmove,iant,ntrials);
  if((iant!=-1 || iant0!=0) && ntrials>=20*na){
    printf("COULD NOT CORRECT CONFIGURATION!!!!!!\n");
    exit(1);
    //return;
  }
  pinstr[indapo]->resetmoved();  
  printf("%ld corrections were made.\n",ncor);

}

int move(APO *vapo, Instrume **pinstr , Obs **pobs,  Transf **ptransf,
	 Site **psite,
	 int napo, int iapo, int iant, float dx, float dy,float dptacc,
	 int sameinstr, int sameobs, int samesite, int samegrid){  

  float x1,y1,phi,x2,y2,dx2,dy2,dpt,dptmax;
  int nconf1,iconf1[3], test=0, ndptacc,i;

  dpt=sqrt(dx*dx+dy*dy);
  if(dpt==0. || (pinstr[iapo]->getfixed(iant))==1){
        pinstr[iapo]->move(iant,0.,0.);
	return 1;
  }
  
  pinstr[iapo]->getant(&x1,&y1,&nconf1,iconf1,iant);
  dptmax=pinstr[iapo]->getdptmax(iant, x1, y1, dx, dy, dptacc,
			   nconf1, iconf1, ptransf[iapo]);
  dptmax*=0.95; 
  if(dptmax<dptacc/4.) return 0;
  
 // printf("dpt=%f dptmax=%f\n",dpt,dptmax);
  if(dpt>dptmax) {dx*=(dptmax/dpt);dy*=(dptmax/dpt);}
  x2=x1+dx; y2=y1+dy;

  if(sameinstr && samesite ) //&& sameobs
    test=vapo[iapo].check(iant,x2,y2,nconf1,iconf1);
  else {
    printf("under developement\n");
    exit(1);
  }
  
  if(test==1){
    phi=atan2(dy,dx);
    ndptacc=(int)(sqrt(dx*dx+dy*dy)/dptacc);
    i=1;
    while(i<ndptacc && test==1){
      x2=x1+dx-i*dptacc*cos(phi); y2=y1+dy-i*dptacc*sin(phi);
      if(sameinstr && samesite ) //&& sameobs
	test=vapo[iapo].check(iant,x2,y2,nconf1,iconf1);
      else {
	printf("under developement\n");
	exit(1);
      }
      if(test==1){
	dx2=dx+i*dptacc*cos(phi); dy2=dy+i*dptacc*sin(phi);
	dpt=sqrt(dx2*dx2+dy2*dy2);
	if(dpt<=dptmax){
	  x2=x1+dx2; y2=y1+dy2;
	  if(sameinstr && samesite ) //&& sameobs
	    test=vapo[iapo].check(iant,x2,y2,nconf1,iconf1);
	  else {
	    printf("under developement\n");
	    exit(1);
	  }
	}
      }
      i++;
    }  
  }
  if(test==0){
    if(!samegrid) for(i=0;i<napo;i++) vapo[i].subtractvisi(iant);
    pinstr[iapo]->move(iant,x2-x1,y2-y1);
    // printf("test=%d move %d dx=%f dy=%f\n",test,iant,x2-x1,y2-y1);
    if(!samegrid) 
      for(i=0;i<napo;i++) {
	vapo[i].addvisi(iant);
	vapo[i].getgrad();
      }
    return 1;
  }
  return 0;
}

float solveFWHM(float B, float Bmax){

  //compute the FWHM for B and Bmax
  // ONLY FOR FWHM IN [0.5,4]*B!!!!

  float a2, x, y, z0, z, y0, yN, dy, FWHM;
  int N, i;
  
  a2=1./(8.*log(2.));
  x=B/(1.44*Bmax);
  y0=0.5; yN=3.;
  N=(int) ((yN-y0)*100.)+1;
  dy=(yN-y0)/(N-1);
  //printf("dy=%f N=%d\n",dy,N);
  y=y0;
  i=0;
  z0=1./(2*a2*y*y)-x*(exp(1./(2*a2*y*y))-1);
  z0*=z0;
  z=0.;
  //printf("z0=%f  z=%f\n",z0,z);
  while(z<z0 && i<N-1){
    i++;
    y+=dy;
    if(i>1) z0=z;
    z=1./(2*a2*y*y)-x*(exp(1./(2*a2*y*y))-1);
    z*=z;
  }
  return (y-dy)*B;
}


float bilinear(float ix,float iy, float **tab){
  int j,k;
  float t,u;

  j=(int) iy; k=(int) ix; 
  t=iy-j; u=ix-k;  
  return ((1.-t)*(1.-u)*tab[j][k]+t*(1.-u)*tab[j+1][k]+
    t*u*tab[j+1][k+1]+(1.-t)*u*tab[j][k+1]);
}

float **matrix(int nrow,int ncol){
  float **m;
  int i,j;
  //  printf("nrow=%d ncol=%d\n",nrow,ncol);
  m=new float*[nrow];
  m[0]=new float[nrow*ncol];
  for(j=1;j<nrow;j++) m[j]=m[j-1]+ncol;
  for(j=0;j<nrow;j++) 
    for(i=0;i<ncol;i++) m[j][i]=0.0;
  return m;
}
void delete_matrix(float **m){
  if(m!=NULL){
    delete m[0];
    delete m;
  }
  else printf("delete_matrix: pointer already NULL\n");
}
int **intmatrix(int nrow,int ncol){
  int **m;
  int i,j;
  //  printf("nrow=%d ncol=%d\n",nrow,ncol);
  m=new int*[nrow];
  m[0]=new int[nrow*ncol];
  for(j=1;j<nrow;j++) m[j]=m[j-1]+ncol;
  for(j=0;j<nrow;j++) 
    for(i=0;i<ncol;i++) m[j][i]=0;
  return m;
}
void delete_intmatrix(int **m){
  if(m!=NULL){
    delete m[0];
    delete m;
  }
  else printf("delete_matrix: pointer already NULL\n");
}

int compi(const void *x, const void *y){
  int z;
  int *p=(int *) x, *q=(int *) y;
  // p=(int *) x; q=(int *) y;
  z=*p-*q;
  return ((z == 0.0) ? 0 : (z < 0.0) ? -1 : +1);
}

int compf(const void *x, const void *y){
  float z;
  float *p=(float *) x, *q=(float *) y;
  // p=(float *) x; q=(float *) y;
  z=*p-*q;
  return ((z == 0.0) ? 0 : (z < 0.0) ? -1 : +1);
}

int compvec(const void *v1, const void *v2){
  float z,r1,r2;
  float *vec1=*((float **) v1), *vec2=*((float **) v2);
  r1=vec1[0]*vec1[0]+vec1[1]*vec1[1];
  r2=vec2[0]*vec2[0]+vec2[1]*vec2[1];

  // p=(float *) x; q=(float *) y;
  z=r1-r2;
  return ((z == 0.0) ? 0 : (z < 0.0) ? -1 : +1);
}

int compdist(const void *v1, const void *v2){

  float z,r1,r2;
  float *vec1=*((float **) v1), *vec2=*((float **) v2);

  r1=vec1[0];
  r2=vec2[0];
  z=r1-r2;
  return ((z == 0.0) ? 0 : (z < 0.0) ? -1 : +1);
}

void printfitserror( int status)
{
    /*****************************************************/
    /* Print out cfitsio error messages and exit program */
    /*****************************************************/

    if (status)
    {
       fits_report_error(stderr, status); /* print error report */

       exit( status );    /* terminate the program, returning error status */
    }
    return;
}


//=======================================================================

Visi::Visi(){ u=0.; v=0.; wght=1.;};
void Visi::set(float u1, float v1, float wght1){
  u=u1; v=v1; wght=wght1;
  r=sqrt(u*u+v*v);
  t=atan2(v,u);
}
void Visi::set(float u1, float v1, float ix1, float iy1,float wght1){
  u=u1; v=v1; ix=ix1; iy=iy1; wght=wght1;
  r=sqrt(u*u+v*v);
  t=atan2(v,u);
}
void Visi::set(float u1, float v1, float r1, float t1,
	   float ix1, float iy1,float wght1){
  u=u1; v=v1; r=r1; t=t1; ix=ix1; iy=iy1; wght=wght1;
}
void Visi::operator=(Visi visi2){
  u=visi2.u; v=visi2.v; ix=visi2.ix; iy=visi2.iy; wght=visi2.wght;
  r=sqrt(u*u+v*v);
  t=atan2(v,u);
}
void Visi::operator=(float x){
  u=x; v=x; ix=0; iy=0; wght=1;   
  r=sqrt(u*u+v*v);
  t=atan2(v,u);
}
void Visi::getijwght(int *p, int *q, float *pwght1){
  *p=(int) ix; *q=(int) iy; *pwght1=wght;
}
void Visi::getuvwght(float *p, float *q, float *pwght1){
  *p=u; *q=v; *pwght1=wght;
}
void Visi::sym(Grid *pgrid){
  u=-u; v=-v; 
  //t = (t>=0.) ? t-PI : t+PI;
  t=atan2(v,u);
  pgrid->getixiy(&ix, &iy, r, t, u, v);
}

//=======================================================================

APO::APO(){

  pvisi=new Visi[NANTVISIMAX];
  dpt=matrix(NATOTMAX+1,3);
  dpt[NATOTMAX]=dpt[0];
  pweights=matrix(NPTSMAX,NCONFMAX);
  pgradx=matrix(MAXDENSITYSIZE,2*MAXDENSITYSIZE);
  pgrady=matrix(MAXDENSITYSIZE,2*MAXDENSITYSIZE);
  pdensity=matrix(MAXDENSITYSIZE,2*MAXDENSITYSIZE);
  pgrid=NULL;
  psite=NULL;
  pinstr=NULL;
  pobs=NULL;
  ptransf=NULL;
}

APO::~APO(){
  delete(pvisi);
  dpt[0]=dpt[NATOTMAX];
  delete_matrix(dpt);
  delete_matrix(pweights);
  delete_matrix(pgradx);
  delete_matrix(pgrady);
  delete_matrix(pdensity);
  }

void APO::set(Instrume *pinstr1, Obs *pobs1, Transf *ptransf1, 
	      Site *psite1, Grid *pgrid1){
  int i,j, test;

  pinstr=pinstr1; pobs=pobs1; psite=psite1; pgrid=pgrid1;
  ptransf=ptransf1;
  ptransf->set(pobs,psite);
  
  initweights(pinstr->nconf);
  reset();
  test=pinstr->check(ptransf);

  if(test==1) {
    printf("\nWARNING from APO creator:\n");
    printf("   Observationnal situation do not match ");
    printf("instrument constraints:\n");
    printf("   Visibilities might be outside the authorized region!!!\n");
    //    exit(1);
  }
  test=pinstr->check(psite);
  if(test==1) {
    printf("\nWARNING from APO creator:\n");
    printf("   Antenna positions do not match ground constraints!!!\n");
    //   exit(1);
  }
}

void APO::set(Grid *pgrid1){
  int i,j,size;

  pgrid=pgrid1;
  size=pgrid->getsize();
  for(j=0;j<=size+1;j++) 
    for(i=0;i<=2*size;i++) pdensity[j][i]=0.0;
  addallvisi();
 }

void APO::reset(){
  int i,j,size;
  size=pgrid->getsize();
  for(j=0;j<=size+1;j++) 
    for(i=0;i<=2*size;i++) pdensity[j][i]=0.0;
  addallvisi();
 }

void APO::correct(float dptacc){
  float dx, dy, dx2, dy2, a, t, randmax;
  int i,iant=0, iant0=0,couldmove,natot;
  long ncor=0l, ntrials=0l;
  couldmove=0;
  natot=pinstr->getnatot();
  randmax=(float) RAND_MAX;
  srand(time(NULL));
  while((iant!=-1 || iant0!=0) && ntrials<20*natot){
    //    printf("ncor=%ld iant=%d\n",ncor,iant);
    if(iant!=-1 && !couldmove) iant0=(iant+1)%natot;
    if(iant==-1 && iant0!=0) iant0=0;
    //printf("natot=%d iant0=%d\n",natot,iant0);
    pinstr->getcorrection(ptransf, psite, iant0, &iant, &dx, &dy);
    //  printf("\n ncor=%ld iant=%d dx=%f dy=%f\n",ncor,iant,dx,dy);
    if(iant!=-1){
      couldmove=move(iant,dx,dy,dptacc);
      i=0;
      while(!couldmove && i<20){
	a=1+rand()*1.2/randmax;
	t=(-45+rand()*90./randmax)*PI/180.;
	dx2=a*dx*cos(t)+a*dy*sin(t);
	dy2=-a*dx*sin(t)+a*dy*cos(t);
	couldmove=move(iant,dx2,dy2,dptacc);
	i++;
      }
      if(couldmove) {
	//printf("moved\n");
	iant0=0;
	printf("\n");
	ncor++;
      }

	//printf("could not move\n");
      
    }
    //else if(ntrials==0) couldmove=1;
    ntrials++;
  }
  //printf("couldmove=%d iant=%d ntrials=%ld\n",couldmove,iant,ntrials);
  if((iant!=-1 || iant0!=0) && ntrials>=20*natot){
    printf("COULD NOT CORRECT CONFIGURATION!!!!!!\n");
    exit(1);
    //return;
  }
  pinstr->resetmoved();
  printf("%ld corrections were made.\n",ncor);
}

int APO::check(int iant, float x, float y,int nconf1,int *iconf1){
 int test;

  test=psite->check(x,y);
  if(test==0) test=pinstr->check(iant,x,y,nconf1,iconf1, ptransf);
  return test;
}

void APO::initweights(int nconf){
  int iconf,isample,nw;
  float norm,tmp,total;
  float coslat, sinlat, cosdec, sindec, cosdl,sindl,cosha;
  
  total=0.0; nw=0;
  ptransf->get(&coslat, &sinlat, &cosdec, &sindec, &cosdl,&sindl);
  norm=pow(1.-exp(-psite->tau/(sinlat*sindec+coslat*cosdec)),2.);
  for(iconf=0;iconf<nconf;iconf++){
    tmp=pobs->vintime[iconf]/(pobs->vnpts[iconf]*pobs->dt);
    printf("tmp=%f\n",tmp);
    for(isample=0;isample < pobs->vnpts[iconf];isample++){
      cosha=ptransf->getcosha(isample,iconf);
      pweights[isample][iconf]=norm*tmp/
	pow(1.-exp(-psite->tau/(sindl+cosdl*cosha)),2);
      total+=pweights[isample][iconf];
      nw++;
     } 
  }
  printf("total=%f nw=%d\n",total,nw);
  for(iconf=0;iconf<nconf;iconf++)
    for(isample=0;isample < pobs->vnpts[iconf];isample++)
      pweights[isample][iconf]*=(nw/total);
}


void APO::getvisi(int iant){

  int i,j,jant,isample,conf;
  float blx, bly, u, v, r, t, ix, iy;
  Ant anti, antj;
  anti=pinstr->getant(iant);
  nvisi=0l; 
  for(i=0;i<anti.nconf;i++){
    conf=anti.iconf[i];
    for(j=0;j<(pinstr->na);j++){
      jant=conf*(pinstr->na-pinstr->nsh)+j;
      if(jant==iant) continue;
      antj=pinstr->AntList[jant];
      blx=anti.x-antj.x;
      bly=anti.y-antj.y;
      for(isample=0;isample<pobs->vnpts[conf];isample++){
	ptransf->xy2uv(blx, bly, conf, isample, &u, &v);
	r=sqrt(u*u+v*v);
	t=atan2(v,u);
	pgrid->getixiy(&ix, &iy, r, t, u, v);
	pvisi[nvisi].set(u,v,r,t,ix,iy,pweights[isample][conf]);
	nvisi++;
      }
    }
  }
  if(nvisi>NANTVISIMAX){
    printf("ERROR: too many visibilities : %ld\n",nvisi);   
    exit(1);
  }
}

void APO::getforce(float *pfu, float *pfv, int iant, float fa){

  int i,j,jant,isample,conf,size, moved, npts1,npts2,na,nconf;
  int nconf1,iconf1[3];
  float x1,y1,x2,y2;
  float daz, dr;
  float blx, bly, u, v, r, t, ix, iy;
  float tmpfaz, tmpfrad,tmpfu, tmpfv, pfu1, pfv1;
  Ant anti, antj;

  *pfu=0.0; *pfv=0.0;
  if((pinstr->getmoved(iant))==1) return ;
  if((pinstr->getfixed(iant))==1) return ;
  na=pinstr->getna();
  nconf=pinstr->getnconf();
  npts1=nvisitot/(nconf*na*(na-1));
  daz=pgrid->getdaz();
  size=pgrid->getsize();
  anti=pinstr->AntList[iant];
  anti.get(&x1,&y1,&nconf1,iconf1);
  for(i=0;i<nconf1;i++) {
    conf=iconf1[i];
    npts2=pobs->getnpts(conf);
    for(j=0;j<(pinstr->na);j++){
      jant=conf*(pinstr->na-pinstr->nsh)+j;
      if(jant==iant) continue;
      antj=pinstr->AntList[jant];
      antj.get(&x2,&y2);
      blx=x1-x2;
      bly=y1-y2;
      for(isample=0;isample<npts2;isample++){
	ptransf->xy2uv(blx, bly, conf, isample, &u, &v);
	r=sqrt(u*u+v*v);
	t=atan2(v,u);
	pgrid->getixiy(&ix, &iy, r, t, u, v);
	if(ix>=0. && iy>=0.){
	  tmpfaz=fa*bilinear(ix,iy+0.5,pgradx)*daz;
	  dr=pgrid->getdr(iy);
	  tmpfrad=fa*bilinear(ix+0.5,iy,pgrady)*dr;

	  tmpfu=tmpfrad*cos(t)+r*cos(t+tmpfaz)-u;
	  tmpfv=tmpfrad*sin(t)+r*sin(t+tmpfaz)-v;

	  ptransf->uv2xy(tmpfu,tmpfv, conf, isample, &pfu1, &pfv1);
	  *pfu+=pfu1*pweights[isample][conf]/(npts1*npts2);
	  *pfv+=pfv1*pweights[isample][conf]/(npts1*npts2);
	}
	else{
	  u=-u; v=-v; 
	  t = (t>0.) ? t-PI : t+PI;
	  pgrid->getixiy(&ix, &iy, r, t, u, v);
	  if(ix>=0. && iy>=0.){
	    tmpfaz=fa*bilinear(ix,iy+0.5,pgradx)*daz;
	    dr=pgrid->getdr(iy);
	    tmpfrad=fa*bilinear(ix+0.5,iy,pgrady)*dr;
	
	    tmpfu=tmpfrad*cos(t)+r*cos(t+tmpfaz)-u;
	    tmpfv=tmpfrad*sin(t)+r*sin(t+tmpfaz)-v;

	    ptransf->uv2xy(tmpfu, tmpfv, conf, isample, &pfu1, &pfv1);
	    *pfu-=pfu1*pweights[isample][conf]/(npts1*npts2);
	    *pfv-=pfv1*pweights[isample][conf]/(npts1*npts2);

	  }
	}
      }
    }
  }
  if(nvisi>NANTVISIMAX){
    printf("ERROR: too many visibilities");   
    exit(1);
  }
  //printf(" pfu=%f pfv%f\n",*pfu,*pfv);
}

void APO::getforce_shad(float *pfu, float *pfv, int iant, float fa, float gsh){

  int i,j,jant,isample,conf,size, moved, npts1,npts2,na,nconf;
  int nconf1,iconf1[3];
  float x1,y1,x2,y2;
  float daz, dr, diam, area;
  float blx, bly, u, v, r, t, ix, iy;
  float tmpfaz, tmpfrad,tmpfu, tmpfv, pfu1, pfv1, fsh;
  Ant anti, antj;

  *pfu=0.0; *pfv=0.0;
  if((moved=pinstr->getmoved(iant))==1) return ;
  na=pinstr->getna();
  nconf=pinstr->getnconf();
  npts1=nvisitot/(nconf*na*(na-1));
  daz=pgrid->getdaz();
  size=pgrid->getsize();
  anti=pinstr->AntList[iant];
  diam=anti.getdiam();
  area=PI*diam*diam/4.;
  anti.get(&x1,&y1,&nconf1,iconf1);
  for(i=0;i<nconf1;i++) {
    conf=iconf1[i];
    npts2=pobs->getnpts(conf);
    for(j=0;j<na;j++){
      jant=conf*(na-pinstr->nsh)+j;
      if(jant==iant) continue;
      antj=pinstr->AntList[jant];
      antj.get(&x2,&y2);
      blx=x1-x2;
      bly=y1-y2;
      for(isample=0;isample<npts2;isample++){
	ptransf->xy2uv(blx, bly, conf, isample, &u, &v);
	r=sqrt(u*u+v*v);
	t=atan2(v,u);
	pgrid->getixiy(&ix, &iy, r, t, u, v);
	//	printf("ix=%f iy=%f \n",ix,iy);
	if(r<diam){
	  fsh=gsh*0.5*(diam*acos(r/diam)-r*sqrt((diam*diam)-(r*r)))/area;
	  // printf("diam%f fsh=%f \n",diam,fsh);
	}
	else fsh=0.;
	if(ix>=0. && iy>=0.) {
	  tmpfaz=fa*bilinear(ix,iy+0.5,pgradx)*daz;
	  dr=pgrid->getdr(iy);
	  tmpfrad=fa*(fsh+bilinear(ix+0.5,iy,pgrady)*dr);

	  tmpfu=tmpfrad*cos(t)+r*cos(t+tmpfaz)-u;
	  tmpfv=tmpfrad*sin(t)+r*sin(t+tmpfaz)-v;

	  ptransf->uv2xy(tmpfu,tmpfv, conf, isample, &pfu1, &pfv1);
	  *pfu+=pfu1*pweights[isample][conf]/(npts1*npts2);
	  *pfv+=pfv1*pweights[isample][conf]/(npts1*npts2);
	  /*
	  if(fsh!=0){
	    printf("tmpfrad=%f faz=%f \n",tmpfrad,
		   sqrt((r*cos(t+tmpfaz)-u)*(r*cos(t+tmpfaz)-u)+
			(r*sin(t+tmpfaz)-v)*(r*sin(t+tmpfaz)-v) ));
	    printf(" pfu=%f pfv=%f fsh=%f \n",*pfu,*pfv, fsh);
	  }
	  */
	}
	else{
	  u=-u; v=-v; 
	  t = (t>0.) ? t-PI : t+PI;
	  pgrid->getixiy(&ix, &iy, r, t, u, v);
	  if(ix>=0. && iy>=0.){
	    tmpfaz=fa*bilinear(ix,iy+0.5,pgradx)*daz;
	    dr=pgrid->getdr(iy);
	    tmpfrad=fa*(fsh+bilinear(ix+0.5,iy,pgrady)*dr);
	
	    tmpfu=tmpfrad*cos(t)+r*cos(t+tmpfaz)-u;
	    tmpfv=tmpfrad*sin(t)+r*sin(t+tmpfaz)-v;

	    ptransf->uv2xy(tmpfu, tmpfv, conf, isample, &pfu1, &pfv1);
	    *pfu-=pfu1*pweights[isample][conf]/(npts1*npts2);
	    *pfv-=pfv1*pweights[isample][conf]/(npts1*npts2);

	  }
	}
      }
    }
  }
  if(nvisi>NANTVISIMAX){
    printf("ERROR: too many visibilities");   
    exit(1);
  }
  /*
  if(fsh!=0){
    
    printf("tmpfu=%f tmpfv=%f tmpfrad=%f faz=%f \n",tmpfu,tmpfv,tmpfrad,
	   sqrt((r*cos(t+tmpfaz)-u)*(r*cos(t+tmpfaz)-u)+
		(r*sin(t+tmpfaz)-v)*(r*sin(t+tmpfaz)-v) ));
    printf(" pfu=%f pfv=%f fsh=%f \n",*pfu,*pfv, fsh);
  }
  */
}
void APO::getgrad(){
  int i,j ,size;

  size=pgrid->getsize();
  i=2*size;
  for(j=0;j<=size;j++) pdensity[j][0]=pdensity[j][i];
  for(i=0;i<=2*size;i++) pdensity[size][i]=pdensity[size-1][i];
  
  for(i=0;i<2*size;i++){
    for(j=0;j<size;j++){
      pgradx[j+1][i]=pdensity[j][i]-pdensity[j][i+1];
      pgrady[j+1][i]=pdensity[j][i]-pdensity[j+1][i];
    }
    pgradx[size+1][i]=pdensity[size][i]-pdensity[size][i+1];
    pgradx[0][i]=0.0;
    pgrady[0][i]=0.0;
  }
  for(j=0;j<size;j++) {
    pgradx[j][i]=pgradx[j][0];
    pgrady[j+1][i]=pdensity[j][i]-pdensity[j+1][i];
  }
  for(j=size;j<=size+1;j++) pgradx[j][i]=pgradx[j][0];
  pgrady[0][i]=0.0;
  i++;
  for(j=0;j<=size;j++) pgrady[j][i]=pgrady[j][1];
}

void APO::addallvisi(){
  int i,j,ivisi,iant, natot;
  float wght;
  natot=pinstr->getnatot();
  nvisitot=0l;
  for(iant=0;iant<natot;iant++){
    getvisi(iant);
    nvisitot+=nvisi;
    for(ivisi=0;ivisi<nvisi;ivisi++){
      pvisi[ivisi].getijwght(&i,&j,&wght);
      if(i >= 0 && j >= 0) pdensity[j][i+1]+=wght;
    }
  }
  getgrad();
}

void APO::writeuvcov(char* filename, long npixwidth, float du){
  int i,j,ivisi,iant, natot,status=0;
  float u,v,wght,crpix[2],crval[2],cdelt[2];
  float bslmax, diam, fwhm;
  double totwght=0;
  float **uvcov;
  long naxes[2];
  char ctype1[MAXSTRING], ctype2[MAXSTRING];
  fitsfile *fptr;  // pointer to the FITS file, defined in fitsio.h 

  bslmax=pinstr->getdmax();
  diam=pinstr->getdiam(0);
  fwhm=pgrid->getfwhm();
  printf("creating fits file %s\n",filename);
  if ( fits_create_file(&fptr, filename, &status) )
    printfitserror( status );
  naxes[0]=npixwidth; 
  naxes[1]=npixwidth;
  if ( fits_create_img(fptr, FLOAT_IMG, 2, naxes, &status) )
    printfitserror( status );
  printf("fits file %s created\n",filename);
  sprintf(ctype1,"%s","Baseline along U");
  if (fits_update_key(fptr,TSTRING, "CTYPE1", ctype1 ,
		      " ",&status) )
    printfitserror( status );
  sprintf(ctype1,"%s","Baseline along V");
  if (fits_update_key(fptr,TSTRING, "CTYPE1", ctype1 ,
		      " ",&status) )
    printfitserror( status );
  crpix[0]=(npixwidth)/2.+1;
  crpix[1]=(npixwidth)/2.+1;
  if (fits_update_key(fptr,TFLOAT, "CRPIX1", &crpix[0],
		      " ",&status) )
    printfitserror( status );
  if (fits_update_key(fptr,TFLOAT, "CRPIX2", &crpix[1],
		      " ",&status) )
    printfitserror( status );
  crval[0]=0.;
  crval[1]=0.;
  if (fits_update_key(fptr,TFLOAT, "CRVAL1", &crval[0],
		      " ",&status) )
    printfitserror( status );
  if (fits_update_key(fptr,TFLOAT, "CRVAL2", &crval[1],
		      " ",&status) )
    printfitserror( status );
  cdelt[0]=du;
  cdelt[1]=du;
  if (fits_update_key(fptr,TFLOAT, "CDELT1", &cdelt[0],
		      " ",&status) )
    printfitserror( status );
  if (fits_update_key(fptr,TFLOAT, "CDELT2", &cdelt[1],
		      " ",&status) )
    printfitserror( status );
  uvcov=matrix(npixwidth,npixwidth);
  natot=pinstr->getnatot();
  for(iant=0;iant<natot;iant++){
    getvisi(iant);
    for(ivisi=0;ivisi<nvisi;ivisi++){
      pvisi[ivisi].getuvwght(&u,&v,&wght);
      i=int (floor(u/du+crpix[0]-.5));
      j=int (floor(v/du+crpix[1]-.5));
      if(i<npixwidth && i>0 && j<npixwidth && j>0){
	uvcov[j][i]+=wght;
        totwght+=wght;
      }
      /* u=-u;
       * v=-v;
       * i=int (floor(u/du+crpix[0]-.5));
       * j=int (floor(v/du+crpix[1]-.5));
       */
      i=int (2*crpix[0]-2-i);
      j=int (2*crpix[1]-2-j);
      if(i<npixwidth && i>0 && j<npixwidth && j>0){
	uvcov[j][i]+=wght;
        totwght+=wght;
      }
    }
  }
  if (fits_update_key(fptr,TDOUBLE, "TOTAL", &totwght,
		      " ",&status) )
    printfitserror( status );
  if (fits_update_key(fptr,TFLOAT, "DIAM", &diam,
		      " ",&status) )
    printfitserror( status );
  if (fits_update_key(fptr,TFLOAT, "BSLMAX", &bslmax,
		      " ",&status) )
    printfitserror( status );
  if (fits_update_key(fptr,TFLOAT, "FWHM", &fwhm,
		      " ",&status) )
    printfitserror( status );
  printf("writing fits file %s\n",filename);
  if ( fits_write_img(fptr, TFLOAT, 1, npixwidth*npixwidth, 
		      uvcov[0], &status) )
    printfitserror( status );
  if ( fits_close_file(fptr, &status) )
    printfitserror( status );
  delete_matrix(uvcov);
}

void APO::addvisi(int iant){
  int i,j,ivisi;
  float wght;
  getvisi(iant);
  for(ivisi=0;ivisi<nvisi;ivisi++){
    pvisi[ivisi].getijwght(&i,&j,&wght);
    if(i >= 0 && j >= 0) pdensity[j][i+1]+=wght;
    else {
      pvisi[ivisi].sym(pgrid);
      pvisi[ivisi].getijwght(&i,&j,&wght);
      if(i >= 0 && j >= 0) pdensity[j][i+1]+=wght;
    }
  }
}

void APO::subtractvisi(int iant){
  int i,j,ivisi;
  float wght;

  getvisi(iant);
  for(ivisi=0;ivisi<nvisi;ivisi++){
    pvisi[ivisi].getijwght(&i,&j,&wght);
    if(i >= 0 && j >= 0) pdensity[j][i+1]-=wght;
    else {
      pvisi[ivisi].sym(pgrid);
      pvisi[ivisi].getijwght(&i,&j,&wght);
      if(i >= 0 && j >= 0) pdensity[j][i+1]-=wght;
    }
  }
}

int APO::move(int iant, float dx, float dy,float dptacc){
  float x1,y1,phi,x2,y2,dx2,dy2,dpt,dptmax;
  int nconf1,iconf1[3], test=0, ndptacc,i;

  dpt=sqrt(dx*dx+dy*dy);
  if(dpt==0.){
        pinstr->move(iant,0.,0.);
	return 1;
  }

  pinstr->getant(&x1,&y1,&nconf1,iconf1,iant);
  dptmax=pinstr->getdptmax(iant, x1, y1, dx, dy, dptacc, nconf1, iconf1, ptransf);
  dptmax*=0.95;
  if(dptmax<dptacc/4.) return 0;
  
  //printf("dpt=%f dptmax=%f\n",dpt,dptmax);
  if(dpt>dptmax) {dx*=(dptmax/dpt);dy*=(dptmax/dpt);}
  x2=x1+dx; y2=y1+dy;
  test=psite->check(x2,y2);
  if(test==0) test=pinstr->check(iant,x2,y2,nconf1,iconf1,ptransf);
  if(test==1){
    phi=atan2(dy,dx);
    ndptacc=(int)(sqrt(dx*dx+dy*dy)/dptacc);
    i=1;
    while(i<ndptacc && test==1){
      x2=x1+dx-i*dptacc*cos(phi); y2=y1+dy-i*dptacc*sin(phi);
      //printf("dx=%f dy=%f\n",x2-x1,y2-y1);
      test=psite->check(x2,y2);
      if(test==0) test=pinstr->check(iant,x2,y2,nconf1,iconf1,ptransf);
      if(test==1){
	dx2=dx+i*dptacc*cos(phi); dy2=dy+i*dptacc*sin(phi);
	dpt=sqrt(dx2*dx2+dy2*dy2);
	//printf("dpt=%f\n",dpt);
	if(dpt<=dptmax){
	  x2=x1+dx2; y2=y1+dy2;
	  test=psite->check(x2,y2);
	  if(test==0) test=pinstr->check(iant,x2,y2,nconf1,iconf1,ptransf);
	}
      }
      i++;
    }  
  }
  if(test==0){
    subtractvisi(iant);
    pinstr->move(iant,x2-x1,y2-y1);
    //printf("test=%d move %d dx=%f dy=%f\n",test,iant,x2-x1,y2-y1);
    addvisi(iant);
    getgrad();
    return 1;
  }
  //printf("test=%d\n",test);
  return 0;
}

void APO::shake(float dptacc){
  int i,iant,natot;
  int couldmove;
  float a,t,randmax,dx,dy;

  printf("start shake\n");
  randmax=(float) RAND_MAX;
  srand(time(NULL));
  natot=pinstr->getnatot();
  for(iant=0;iant < natot;iant++) {
    couldmove=0;
    i=0;
    while(!couldmove && i<20){
      a=(1+rand()*1.6/randmax)*dptacc;
      t=(-180+rand()*360./randmax)*PI/180.;
      dx=a*cos(t); dy=a*sin(t);
      couldmove=move(iant,dx,dy,dptacc);
      if(couldmove) printf("\n");
      i++;
    }
  }
  pinstr->resetmoved();
}

void APO::optimize(float fa, float dptacc){

  
  int iant,nmoved,k,natot,nok,ind;
  int couldmove;
  float tmpfu, tmpfv, dpt1;
  char pause;
  printf("startoptimize\n");
  natot=pinstr->getnatot();
  if(natot>=NATOTMAX){
    printf("ERROR in APO::optimize : too many antennas!!!");
    exit(1);
  }
  nmoved=0;
  nok=1;
  while(nmoved < natot && nok!=0) {
    nok=0;
    for(iant=0;iant<natot;iant++){
      getforce(&tmpfu, &tmpfv, iant,fa);
      dpt1=tmpfu*tmpfu+tmpfv*tmpfv;
      if(dpt1>0.0){
	dpt[nok][0]=tmpfu; 
	dpt[nok][1]=tmpfv;
	dpt[nok][2]=iant;
      	nok++;
      }
    }
    printf("nok=%d\n",nok);
    qsort(dpt,nok,sizeof(float *),compvec);
    couldmove=0; k=nok-1;
    while(!couldmove && k>=0 && nmoved<natot){ 
      tmpfu=dpt[k][0]; tmpfv=dpt[k][1]; ind=(int) dpt[k][2];
      couldmove=move(ind,tmpfu,tmpfv,dptacc);
      if(!couldmove && k==0) {
	dpt1=sqrt(tmpfu*tmpfu+tmpfv*tmpfv);
	tmpfu=-tmpfu*2.1*dptacc/dpt1; 
	tmpfv=-tmpfv*2.1*dptacc/dpt1; 
	couldmove=move(ind,tmpfu,tmpfv,dptacc);
	if(!couldmove) {
	  move(ind,0.,0.,dptacc);
	  nmoved++; k+=2;
	}
      }
      k--;
      if(k==nok) nmoved=natot;
    }
    printf("k=%d \n",k+1);				
    if(couldmove) nmoved++;
  }
  pinstr->resetmoved();
}   
void APO::optimize(float fa, float dptacc, int conf){

  int i,k,na,nsh,nok,iant, nmoved,ind;
  int couldmove;
  float tmpfu, tmpfv, dpt1;

  printf("startoptimize\n");
  na=pinstr->getna();
  nsh=pinstr->getnsh();
  nmoved=0;
  nok=1;
  while(nmoved < na && nok!=0 ) {
    nok=0;
    for(i=0;i<na;i++){
      iant=conf*(na-nsh)+i;
      getforce(&tmpfu, &tmpfv, iant,fa);
      dpt1=tmpfu*tmpfu+tmpfv*tmpfv;
      if(dpt1>0.0){
	dpt[nok][0]=tmpfu; 
	dpt[nok][1]=tmpfv;
	dpt[nok][2]=iant;
      	nok++;
      }
    }
    printf("nok=%d\n",nok);
    qsort(dpt,nok,sizeof(float *),compvec);
    couldmove=0; k=nok-1;
    while(!couldmove && k>=0 && nmoved<na){ 
      tmpfu=dpt[k][0]; tmpfv=dpt[k][1]; ind=(int) dpt[k][2];
      couldmove=move(ind,tmpfu,tmpfv,dptacc);
      if(!couldmove && k==0) {
	dpt1=sqrt(tmpfu*tmpfu+tmpfv*tmpfv);
	tmpfu=-tmpfu*2.1*dptacc/dpt1; 
	tmpfv=-tmpfv*2.1*dptacc/dpt1; 
	couldmove=move(ind,tmpfu,tmpfv,dptacc);
	if(!couldmove) {
	  move(ind,0.,0.,dptacc);
	  nmoved++; k+=2;
	}
      }
      k--;
      if(k==nok) nmoved=na;
    }
    printf("k=%d \n",k+1);				
    if(couldmove) nmoved++; 
  }
  pinstr->resetmoved();
}   


float APO::getstdev(){
  float stdev,integr;
  int i,j,size;

  size=pgrid->getsize();
  integr=pgrid->getintegr();
  integr*=nvisitot;
  stdev=0.0;
  for(j=0;j<size;j++) 
    for(i=0;i<2*size;i++) 
      stdev+=((pdensity[j][i]-integr)*(pdensity[j][i]-integr));
  return sqrt(stdev/(size*size*2));
}
void APO::printvisi(){
  int i;
  printf("\n-------------------------------\n");
  printf("        PRINT VISIBILITIES\n");
  printf("-------------------------------\n");
  printf("nvisi= %d \n",nvisi);
  for(i=0;i<nvisi;i++) pvisi[i].print();
  printf("-------------------------------\n");
}

void APO::printdensity(){
  int i,j,size;
  if(pgrid==0) size=MAXDENSITYSIZE;
  else size=pgrid->getsize();
  printf("\n-------------------------------\n");
  printf("        PRINT DENSITY\n");
  printf("-------------------------------\n");
  printf("size= %d \n",size);
  for(j=0;j<=size;j++) {
    for(i=0;i<=2*size;i++) printf("%4.1f ",pdensity[j][i]);
    printf("\n");
  }
  printf("-------------------------------\n");
  for(j=0;j<=size+1;j++) {
    for(i=0;i<=2*size;i++) printf("%4.1f ",pgradx[j][i]);
    printf("\n");
  }
  printf("-------------------------------\n");
  for(j=0;j<=size;j++) {
    for(i=0;i<=2*size+1;i++) printf("%4.1f ",pgrady[j][i]);
    printf("\n");
  }
  printf("-------------------------------\n");
}
void APO::printforce(int iant,float fa,char *name){

  int i,j,jant,isample,conf,size, moved, npts1,npts2,na,nconf;
  int nconf1,iconf1[3];
  float x1,y1,x2,y2;
  float daz, dr;
  float cosha, sinha,pfu,pfv;
  float blx, bly, u, v, r, t, ix, iy, tmpfaz, tmpfrad,tmpfu, tmpfv;
  Ant anti, antj;
  FILE *fp;

  if((fp=fopen(name,"w")) == NULL){
    printf("ERROR: cannot open file %s\n",name);
    exit(1);
  }

  if((moved=pinstr->getmoved(iant))==1) return ;
  na=pinstr->getna();
  nconf=pinstr->getnconf();
  npts1=nvisitot/(nconf*na*(na-1));
  daz=pgrid->getdaz();
  size=pgrid->getsize();
  anti=pinstr->AntList[iant];
  anti.get(&x1,&y1,&nconf1,iconf1);
  for(i=0;i<nconf1;i++) {
    conf=iconf1[i];
    //    npts2=pobs->vnpts[conf];
    npts2=pobs->getnpts(conf);
    for(j=0;j<(na);j++){
      jant=conf*(na-pinstr->nsh)+j;
      if(jant==iant) continue;
      antj=pinstr->AntList[jant];
      antj.get(&x2,&y2);
      blx=x1-x2;
      bly=y1-y2;
      for(isample=0;isample<npts2;isample++){
	ptransf->xy2uv(blx, bly, conf, isample, &u, &v);
	r=sqrt(u*u+v*v);
	t=atan2(v,u);
	pgrid->getixiy(&ix, &iy, r, t, u, v);
	if(ix>=0. && iy>=0.){
	  tmpfaz=fa*bilinear(ix,iy+0.5,pgradx)*daz;
	  dr=pgrid->getdr(iy);
	  tmpfrad=fa*bilinear(ix+0.5,iy,pgrady)*dr;

	  tmpfu=tmpfrad*cos(t)+r*cos(t+tmpfaz)-u;
	  tmpfv=tmpfrad*sin(t)+r*sin(t+tmpfaz)-v;

	  ptransf->uv2xy(tmpfu,tmpfv, conf, isample, &pfu, &pfv);
	  pfu=pfu*pweights[isample][conf]/(npts1*npts2);
	  pfv=pfv*pweights[isample][conf]/(npts1*npts2);

	  fprintf(fp,"%-7.2f %-7.2f %-7.2f %-7.2f\n",u,v,pfu,pfv);
	}
	else{
	  u=-u; v=-v; 
	  t = (t>0.) ? t-PI : t+PI;
	  pgrid->getixiy(&ix, &iy, r, t, u, v);
	  if(ix>=0. && iy>=0.){
	    tmpfaz=fa*bilinear(ix,iy+0.5,pgradx)*daz;
	    dr=pgrid->getdr(iy);
	    tmpfrad=fa*bilinear(ix+0.5,iy,pgrady)*dr;
	
	    tmpfu=tmpfrad*cos(t)+r*cos(t+tmpfaz)-u;
	    tmpfv=tmpfrad*sin(t)+r*sin(t+tmpfaz)-v;

	    ptransf->uv2xy(tmpfu,tmpfv, conf, isample, &pfu, &pfv);
	    pfu=-pfu*pweights[isample][conf]/(npts1*npts2);
	    pfv=-pfv*pweights[isample][conf]/(npts1*npts2);

	    fprintf(fp,"%-7.2f %-7.2f %-7.2f %-7.2f\n",u,v,pfu,pfv);
	  }
	}
      }
    }
  }
  if(nvisi>NANTVISIMAX){
    printf("ERROR: too many visibilities"); 
    fclose(fp);
    exit(1);
  }
  fclose(fp);
  //printf(" pfu=%f pfv%f\n",*pfu,*pfv);
}
//=======================================================================

int Ant::check(Site *psite){
    return psite->check(x,y);
}
void Ant::offset(float xoff, float yoff){

    x+=xoff;
    y+=yoff;
}
//void Ant::random(Site *psite){
//  psite->random(&x,&y);
//}

//=======================================================================

void Instrume::set(float vx[], float vy[], int vfixed[], float vdiam[], 
		   int na1, int nsh1, int nconf1, float dspec1,
		   float vdmax1[], float vdmin1[]){
  int anconf,i;
  int aconf[3];
  float x1,y1;
  int nconf2,iconf2[3],test;

  nmoved=0; notmoved=0;
  na=na1; nsh=nsh1; nconf=nconf1; 
  dspec=dspec1;
  natot=nconf*na-nsh*(nconf-1);
  AntList=new Ant[natot];
  vdmax=new float[nconf];
  vdmin=new float[nconf];
  for(i=0; i<natot; i++){
    anconf=1;
    if((i >= (na-nsh)) && (i < (nconf-1)*(na-nsh)+nsh) && 
       ((i%(na-nsh)) < nsh)) anconf=2;
    if( nconf>2 && (i >= 2*(na-nsh)-1) && (2*nsh-na >= 0) && 
	((i%(2*(na-nsh))) <= 2*nsh-na) && 
	(i <(nconf-3)*(na-nsh)+na) && (i>=2*(na-nsh)) ) anconf=3;
    for(int j=anconf-1; j>=0;j--) {
      if(i < (nconf-1)*(na-nsh)) aconf[anconf-j-1]=i/(na-nsh)-j;
      else aconf[anconf-j-1]=nconf-1-j;
    }
    AntList[i].set(vx[i],vy[i],vfixed[i],vdiam[i],anconf,aconf);
  }
  qsort(vdiam,natot,sizeof(float),compf);
  ndiam=1;
  diamlist[0]=vdiam[0];
  for(i=1; i<natot; i++)
    if(vdiam[i]!=vdiam[i-1]) {
      ndiam++;
      if(ndiam>NDIAMMAX){
	printf("ERROR: too many different antennas:");
	printf("       the number of diameters should be < %d\n",NDIAMMAX);
	exit(1);
      }
      diamlist[ndiam-1]=vdiam[i];
    }
  printf("Number of different diameters: %d\n",ndiam);
  for (i=0; i<nconf; i++){
    vdmax[i]=vdmax1[i];
    vdmin[i]=vdmin1[i];  
  }
  i=0; test=0;
  while(test=0 && i<natot){
    AntList[i].get(&x1,&y1,&nconf2,iconf2);
    test=check(i,x1,y1,nconf2,iconf2);
    i++;
  }
  if(test==1){
    printf("\nERROR in Insrume creator:\n");
    printf("  There are errors in antenna positions: \n");
    printf("   some antennas are too near to  each other!!!\n");
  }
}

void Instrume::set(float *xx, float *yy, int na){

  float x1,y1;
  int i,nconf2,iconf2[3],test;

  if (na>natot){
    printf("ERROR not so many antennas!\n");
    exit(1);
  }
  for(i=0;i<na;i++) AntList[i].set(xx[i],yy[i]);

  i=0; test=0;
  while(test=0 && i<natot){
    AntList[i].get(&x1,&y1,&nconf2,iconf2);
    test=check(i,x1,y1,nconf2,iconf2);
    i++;
  }
  if(test==1){
    printf("\nERROR in Insrume creator:\n");
    printf("  There are errors in antenna positions: \n");
    printf("   some antennas are too near to  each other!!!\n");
  }
}

void Instrume::setvdmax(float *vdmax1, int nconf1){

  float x1,y1;
  int i,nconf2,iconf2[3],test;

  if(nconf1!=nconf){
    printf("ERROR nconf does not match!\n");
    exit(1);
  }
  for (i=0; i<nconf; i++) vdmax[i]=vdmax1[i];
  i=0; test=0;
  while(test=0 && i<natot){
    AntList[i].get(&x1,&y1,&nconf2,iconf2);
    test=check(i,x1,y1,nconf2,iconf2);
    i++;
  }
  if(test==1){
    printf("\nERROR in Insrume creator:\n");
    printf("  There are errors in antenna positions: \n");
    printf("   some antennas are too near to  each other!!!\n");
  }
}
void Instrume::random(Site *psite, Obs *pobs, Model *pmod){

  int iant,i;
  float u1,u2,umax, norm,fwhm,dec,lat;
  float sig, sig2, rad, rad2;
  float xx, yy, tmp;
  int nconf, iconf[3];
  
  fwhm=pmod->getfwhm();
  dec=pobs->getdec();
  lat=psite->getlat();

  sig=fwhm/(4*sqrt(log(2)));
  sig2=sig*sig;
  umax=RAND_MAX*1.001;
  srand(time(NULL));

  for(iant=0;iant<natot;iant++){
    AntList[iant].get(&tmp, &tmp, &nconf, iconf);
    rad=vdmax[iconf[0]]/2;
    rad2=rad*rad;
    u1=rand()/umax;
    u2=rand()/umax;
    u1=u1*(1-exp(-rad2/(2*sig2)))+exp(-rad2/(2*sig2));
    norm=2*PI*sig2*(1-exp(-rad2/(2*sig2)));
    xx=sig*sqrt(-2*log(u1))*cos(norm*u2);
    yy=sig*sqrt(-2*log(u1))*sin(norm*u2);
    AntList[iant].set(xx,yy);

  }
}


void Instrume::sort(){

  int iant;
  float **posi;
  float *posi0;
  float x,y;
  
  posi=matrix(natot,2);
  posi0=posi[0];
  for(iant=0;iant<natot;iant++){
    AntList[iant].get(&x,&y);
    posi[iant][0]=x; 
    posi[iant][1]=y;
  }
  qsort(posi,natot,sizeof(float *),compvec);
  //qsort(vdmax2,nconf,sizeof(float),compf);
  for(iant=0;iant<natot;iant++)
    AntList[iant].set(posi[iant][0],posi[iant][1]);
  posi[0]=posi0;
  delete_matrix(posi);
}

void Instrume::offset(float xoff, float yoff){
  int i;
  for(i=0;i<natot;i++) AntList[i].offset(xoff, yoff);
}

void Instrume::get(float *xx, float *yy, int *vfixed, float *vdiam, 
		   int *na1, int *nsh1, int *nconf1, float *dspec1,
		   float *vdmax1, float *vdmin1){
  int i;

  *na1=na; *nsh1=nsh; *nconf1=nconf; *dspec1=dspec;
  for(i=0;i<natot;i++) AntList[i].get(xx+i,yy+i,vfixed+i,vdiam+i);
  for(i=0;i<nconf;i++) {
    vdmax1[i]=vdmax[i];
    vdmin1[i]=vdmin[i];
  }
}
void Instrume::getpositions(float *xx, float *yy){
  int i;
  for(i=0;i<natot;i++) AntList[i].get(xx+i,yy+i);
}

void Instrume::getcenter(float *xc, float *yc){
  int i;
  float x, y;
  *xc=0;
  *yc=0;
  for(i=0;i<natot;i++) {
    AntList[i].get(&x, &y);
    (*xc)+=x;
    (*yc)+=y;
  }
  (*xc)/=natot;
  (*yc)/=natot;
}

float Instrume::getdiam(int i){
  if(i>=ndiam) {
    printf("ERROR in Instrume::getdiam : not so many diameters");
    exit(1);
  }
  return diamlist[i];
}

float Instrume::getdmax(){
  float vdmax2[nconf];
  int i;
  
  for(i=0;i<nconf;i++) vdmax2[i]=vdmax[i];
  qsort(vdmax2,nconf,sizeof(float),compf);
  return vdmax2[nconf-1];
}    

float Instrume::getdptmax(int iant,float x1, float y1, float dx ,float dy,
			 float dptacc,int nconf1,int iconf1[3],Transf *ptransf){
  float blx,bly,dmax,dpt,dptmax,dptmax1,ruvmax;
  float x2, y2, dd, dd0;
  int i,j,k,jant,conf,nconf2,iconf2[3];

  dpt=sqrt(dx*dx+dy*dy);
  dx/=dpt; dy/=dpt;
  conf=iconf1[0];
  dmax=vdmax[conf];
  dptmax=dmax/10.;
  for(j=0;j<na;j++){
    jant=conf*(na-nsh)+j;
    if(jant==iant) continue;
    AntList[jant].get(&x2,&y2,&nconf2,iconf2);
    if(iconf2[0] != conf) continue;
    blx=x1-x2;
    bly=y1-y2;
    ruvmax=ptransf->getruvmax(blx+dx*dptmax, bly+dy*dptmax, conf);
    //printf("ruvmax=%f dmax=%f\n",ruvmax,dmax);
    if(ruvmax>dmax){
      dd0=dd=dptmax;
      dptmax1=0.;
      while(dd0>dptacc/2){
	dd0/=2.;
	if(ruvmax>dmax) dd-=dd0;
	else {
	  dptmax1=dd;
	  dd+=dd0;
	}
	ruvmax=ptransf->getruvmax(blx+dx*dd, bly+dy*dd, conf);
      }
      if(ruvmax<dmax) dptmax1=dd;
      if(dptmax1==0.) return 0.;
      dptmax=dptmax1;
    }
  }
  //  printf("dptmax=%f\n",dptmax);
  return dptmax;
}

int Instrume::getBSL_FWHM_ObsDur(float obsdurmax, char *filename, int iconf, 
	      float *baseline, float *fwhm, float *obsdur){

  /*
    tab[*][0]=baseline
    tab[*][1]=alpha
    tab[*][2]=beta
    tab[*][3]=gamma
    tab[*][4]=dtheta
    tab[*][5]=lambda
  */
  int i,j=0;
  char c, pause;
  float **tab;
  float *tab0;
  float diam, Bmax, B0, beta2, bgamma, chi,a;
  double Cst=2.*sqrt(log(10.)/10.*log(2.))/PI;
  FILE *fp;

  if((fp=fopen(filename,"r")) == NULL){
    printf("ERROR: cannot open file %s\n",filename);
    exit(1);
  }
  while((c=getc(fp))!='\n' && c!=EOF);
  if(c==EOF) return 0;
  tab=matrix(30,6);
  tab0=tab[0];
  while(c!=EOF){
    for(i=1;i<6;i++) fscanf(fp,"%f",&tab[j][i]);
    //printf("tab[j][0]=%f\n",tab[j][1]);
    //scanf("%c",&pause);
    while((c=getc(fp))!='\n' && c!=EOF);
    if(tab[j][1]!= 0) {
      tab[j][4]/=RAD2ASEC;
      tab[j][5]*=1.e-3;
      tab[j][0]=Cst*sqrt(tab[j][3])/tab[j][2]* tab[j][5]/(tab[j][4]);
      j++;
    }
  }
  fclose(fp);
  if(iconf>=j){
    tab[0]=tab0;
    delete_matrix(tab);
    return 0;
  }
  qsort(tab,j,sizeof(float *),compvec);
  *baseline=tab[iconf][0];
  //printf("alpha=%f beta=%f  gamma=%f dtheta=%f lambda=%f\n",
  //   tab[iconf][1],tab[iconf][2],tab[iconf][3],tab[iconf][4],tab[iconf][5]);
  diam=getdiam(0);
  Bmax=na*(na-1)*tab[iconf][1]*diam*obsdurmax/(48.*1.44);
  beta2=tab[iconf][2]*tab[iconf][2];
  bgamma=log(10.)*tab[iconf][3]/10.;
  chi=beta2/bgamma*(exp(bgamma/beta2)-1);
  B0=Bmax*1.44/chi;
  printf("\nBmax=%7.1f B0=%7.1f baseline=%7.1f\n",Bmax,B0,*baseline);
  if(*baseline<=B0){
    printf("CLEAN BEAM FEASIBLE: baseline < B0\n");
    *obsdur=48.*chi*(*baseline)/(na*(na-1)*tab[iconf][1]*diam);
    a=1./(2.*sqrt(2*log(2)));
    *fwhm=tab[iconf][2]*(*baseline)/(a*sqrt(2*bgamma));
    //printf("  observation duration required=%-5.2f hours\n",*obsdur);
    //printf("  FWHM=%-8.1f m\n\n",*fwhm);
  }
  else{
    if(*baseline>Bmax){
      printf("WARNING: to achive this resolution(%f asec) with the\n",
	     tab[iconf][4]*RAD2ASEC);
      printf("         sampling specified (alpha=%f)\n",tab[iconf][1]);
      printf("         MULTICONFIGURATION OBSERVATION is required\n");
      *obsdur=obsdurmax;
      *fwhm=2*(*baseline);
      printf("  FWHM=%-7.1f m\n\n",*fwhm);
    }
    else {
      printf("RESOLUTION ATTAINABLE WITH A LOSS OF SENSITIVITY: ");
      printf("baseline > B0\n");
      *fwhm=solveFWHM(*baseline,Bmax);
      *obsdur=obsdurmax;
      printf("  FWHM=%7.1f m\n\n",*fwhm);
    }
  }
  tab[0]=tab0;
  delete_matrix(tab);
  return 1;
}


int Instrume::check(int iant, float x1, float y1, int nconf1,int iconf1[3],
		    Transf *ptransf) {
  float dist,blx,bly;
  float x2, y2, ruvmin;
  int nconf2, iconf2[3];
  int i,j,k,jant,indconf;
  int ncom, icom[3];

  for(jant=0;jant<natot;jant++){
    if(jant==iant) continue;
    ncom=0;
    icom[0]=0;
    AntList[jant].get(&x2,&y2,&nconf2,iconf2);
    blx=(x1-x2); bly=(y1-y2);
    dist=sqrt(blx*blx + bly*bly);
    if(dist<=dspec)  return 1;
    for(j=0;j<nconf1;j++)
      for(k=0;k<nconf2;k++)
	if(iconf2[k]==iconf1[j]){
	  icom[ncom]=iconf1[j]; 
	  ncom++; 
	}
    if(ncom==0) continue;
    indconf=icom[0]; 
    ruvmin=ptransf->getruvmin(blx, bly, indconf);
    //printf("ruvmin=%f dmin=%f\n",ruvmin,vdmin[indconf]);
    if(ruvmin<vdmin[indconf])  return 1;   
  }
  return 0;
}

int Instrume::check(int iant, float x1, float y1, int nconf1,int iconf1[3]){
  float dist, blx, bly, dmin;
  float x2, y2;
  int nconf2, iconf2[3];
  int i,j,k,jant,indconf;
  int ncom, icom[3];

  for(jant=0;jant<natot;jant++){
    if(jant==iant) continue;
    ncom=0;
    icom[0]=0;
    AntList[jant].get(&x2,&y2,&nconf2,iconf2);
    for(j=0;j<nconf1;j++)
      for(k=0;k<nconf2;k++)
	if(iconf2[k]==iconf1[j]){
	  icom[ncom]=iconf1[j]; 
	  ncom++; 
	}
    blx=(x1-x2); bly=(y1-y2);
    dist=sqrt(blx*blx + bly*bly);
    dmin= (ncom==0) ? dspec : vdmin[icom[0]];
      //    indconf = (ncom=0) ? 0 : icom[0];
    //printf("dmin=%f iant=%d jant=%d \n",dmin,iant,jant);
    if(dist<dmin) return 1;
  }
  return 0;
}
int Instrume::check(Transf *ptransf){
  int i, test,nconf,iconf[3];
  float x,y;

  for(i=0;i<natot;i++){
    AntList[i].get(&x,&y,&nconf,iconf);
    test=check(i,x,y,nconf,iconf,ptransf);
    if(test==1) return 1;
  }
  return 0;
}

int Instrume::check(Site *psite){
  int i, test;
  for(i=0;i<natot;i++){
    test=AntList[i].check(psite);
    if(test==1) return 1;
  }
  return 0;
}
void Instrume::move(int iant, float dx, float dy){
  nmoved++;
  AntList[iant].move(dx,dy);
  if(dx==0. && dy==0.) notmoved++;
  else printf("nmoved=%3d iant=%3d dx=%5.1f dy=%5.1f ",nmoved,iant,dx,dy);
}
void Instrume::resetmoved(){
  int i;
  for(i=0;i<natot;i++) AntList[i].resetmoved();
  printf("%5.1f %c of the antennas could not move\n",
	 notmoved*100.0/natot,'%');
  notmoved=0;
  nmoved=0;
}
void Instrume::freeants(){
  int i;
  for(i=0;i<natot;i++) AntList[i].freeant();
}


void Instrume::getcorrection(Transf *ptransf, Site *psite, int iant0,
			     int *iant1, float *dx, float *dy){ 
                             //return iant=-1 if no correction needed
  int i, j, k, iant, jant, iconfmin, iconfmax, test;
  int nconf1,iconf1[3],nconf2, iconf2[3];
  float x1,y1,x2, y2;
  float dist, blx, bly, dmin,dmax;
  int ncom, icom[3];
  float ruvmin, ruvmax;

  //  for(iant=0;iant<natot;iant++){
  //printf("iant0=%d\n",iant0);

  iant=iant0-1;
  *iant1=-1;
  while(*iant1==-1 && iant<natot-1){
    iant++;
    if((getfixed(iant))==1) continue;
    AntList[iant].get(&x1,&y1,&nconf1,iconf1);
    test=AntList[iant].check(psite);
    iconfmax=iconf1[0];
    dmax=vdmax[iconfmax];
    jant=-1;
    while(*iant1==-1 && jant<natot-1){
    //   for(jant=0;jant<natot;jant++){
      jant++;
      if(jant==iant) continue;
      ncom=0;
      icom[0]=0;
      AntList[jant].get(&x2,&y2,&nconf2,iconf2);

      blx=x1-x2; bly=y1-y2;
      dist=sqrt(blx*blx+bly*bly);
      if(iconf2[0]==iconf1[0]) ruvmax=ptransf->getruvmax(blx,bly,iconfmax);
      else ruvmax=0.;
      if(test==1 || ruvmax>dmax){
	if(test!=1)printf("ruv too big ruvmax=%f dmax=%f iant=%d jant=%d\n",
			  ruvmax,dmax, iant, jant);
	else printf("1 forbidden position for iant=%d\n",iant);
	*iant1=iant;
	*dx=-blx*(dmax/5.)/dist;
	*dy=-bly*(dmax/5.)/dist;
      }
      else {
	for(j=0;j<nconf1;j++)
	  for(k=0;k<nconf2;k++)
	    if(iconf2[k]==iconf1[j]){
	      icom[ncom]=iconf1[j]; 
	      ncom++; 
	    }
	if(ncom==0) continue;
	iconfmin=icom[0];
	dmin=vdmin[iconfmin];
	ruvmin=ptransf->getruvmin(blx,bly,iconfmin);
	if(ruvmin<dmin){
	  printf("ruv too small dmin=%f ruvmin=%f\n",dmin,ruvmin);
	  *iant1=iant;
	  *dx=blx*(dmin/2.)/dist;
	  *dy=bly*(dmin/2.)/dist;
	}
      }
    }
  }
}

void Instrume::getcorrection(Transf *ptransf, Site *psite, 
			     int iant0, int conf, 
			     int *iant1, float *dx, float *dy){ 
                             //return iant=-1 if no correction needed
  int i, j, k, iant, jant, iconfmin, iconfmax, test;
  int nconf1,iconf1[3],nconf2, iconf2[3];
  float x1,y1,x2, y2;
  float dist, blx, bly, dmin,dmax;
  int ncom, icom[3];
  float ruvmin, ruvmax;

  //  for(iant=0;iant<natot;iant++){

  iant=conf*(na-nsh)+iant0-1;
  *iant1=-1;
  while(*iant1==-1 && iant<(conf*(na-nsh)+na-1)){
    iant++;
    if((getfixed(iant))==1) continue;
    AntList[iant].get(&x1,&y1,&nconf1,iconf1);
    test=AntList[iant].check(psite);
    iconfmax=iconf1[0];
    dmax=vdmax[iconfmax];
    jant=-1;
    while(*iant1==-1 && jant<(conf*(na-nsh)+na-1)){
    //   for(jant=0;jant<natot;jant++){
      jant++;
      if(jant==iant) continue;
      ncom=0;
      icom[0]=0;
      AntList[jant].get(&x2,&y2,&nconf2,iconf2);

      blx=x1-x2; bly=y1-y2;
      dist=sqrt(blx*blx+bly*bly);
      if(iconf2[0]==iconf1[0]) ruvmax=ptransf->getruvmax(blx,bly,iconfmax);
      else ruvmax=0.;
      if(test==1 || ruvmax>dmax){
	if(test!=1)printf("ruv too big ruvmax=%f dmax=%f iant=%d jant=%d\n",
			  ruvmax,dmax, iant, jant);
	else printf("2 forbidden position for iant=%d\n",iant);
	*iant1=iant;
	*dx=-blx*(dmax/5.)/dist;
	*dy=-bly*(dmax/5.)/dist;
      }
      else {
	for(j=0;j<nconf1;j++)
	  for(k=0;k<nconf2;k++)
	    if(iconf2[k]==iconf1[j]){
	      icom[ncom]=iconf1[j]; 
	      ncom++; 
	    }
	if(ncom==0) continue;
	iconfmin=icom[0];
	dmin=vdmin[iconfmin];
	ruvmin=ptransf->getruvmin(blx,bly,iconfmin);
	if(ruvmin<dmin){
	  printf("ruv too small dmin=%f ruvmin=%f\n",dmin,ruvmin);
	  *iant1=iant;
	  *dx=blx*(dmin/2.)/dist;
	  *dy=bly*(dmin/2.)/dist;
	}
      }
    }
  }
}


void Instrume::addants2pads(Padslist *pads, int iconf){

  int i, fixed, nadded=0;
  float x, y;
  for(i=0;i<natot;i++){
    AntList[i].get(&x, &y, &fixed);
    if (!fixed) {
      pads->add(x,y, iconf);
      nadded++;
    }
  }
  printf("%d antenna positions were added to %s\n",nadded,pads->getfilename());
}

int Instrume::loadants2pads(Padslist *pads, int nloadmax, float dmax, 
			    float ff, int iconf){

  int iant, j, ipad, fixed, nloaded=0, loaded=0, nok=0;
  float x, y, d, xpad, ypad, dist, dist2pad,r, xc, yc;
  float **tmp, *tmp0;

  tmp=matrix(natot,5);
  tmp0=tmp[0];
  getcenter(&xc, &yc);
  for(j=0;j<nloadmax;j++){
    nok=0;
    for(iant=0;iant<natot;iant++){
      AntList[iant].get(&x, &y, &fixed);
      r=sqrt((x-xc)*(x-xc)+(y-yc)*(y-yc));
      if(r<=dmax/10) dist=0.1*ff*dmax*sqrt(PI/natot)/2;
      else dist=ff*r*sqrt(PI/natot)/2;
      //printf("distance to pads has to be < %f m\n",dist);
      if(!fixed){
	dist2pad=pads->getdist2pad(x,y,iconf,&xpad,&ypad,&ipad);
	if(dist2pad>=0 && dist2pad<dist){
	  tmp[nok][0]=dist2pad;
	  tmp[nok][1]=xpad; tmp[nok][2]=ypad;
	  tmp[nok][3]=ipad;
	  tmp[nok][4]=iant;
	  nok++;
	}
      }
    }
    if(nok==0) {
      tmp[0]=tmp0;
      delete_matrix(tmp);
      printf("%d antennas were loaded on pads from %s\n",
	     nloaded,pads->getfilename());
      return nloaded;
    }
    qsort(tmp,nok,sizeof(float *),compdist);
    loaded=pads->load((int) tmp[0][3], iconf);
    if(loaded) {
      AntList[(int) tmp[0][4] ].fixant(tmp[0][1],tmp[0][2]);
      nloaded++;
    }
  }
  tmp[0]=tmp0;
  delete_matrix(tmp);
  printf("%d antennas were loaded on pads from %s\n",nloaded,pads->getfilename());
  return nloaded;
}

/*
  loaded=pads->load(x, y, dist, iconf, &xpad, &ypad);
  if(loaded) {
  AntList[i].fixant(xpad,ypad);
  nloaded++;
  if(nloaded==nloadmax){
  printf("%d antennas were loaded on pads from %s\n",
  nloaded,pads->getfilename());
  return nloaded;
  }
  }
  }
  }
  printf("%d antennas were loaded on pads from %s\n",nloaded,pads->getfilename());
  return nloaded;
  }
  */

void Instrume::print(){
  int i;
  printf("\n-------------------------------\n");
  printf("        PRINT INSTRUMENT\n");
  printf("-------------------------------\n");
  printf("natot= %d \n",natot);
  printf("nconf= %d \n",nconf);
  for(i=0; i<nconf; i++) 
    printf("conf %d dmax=%7.1f dmin=%7.1f \n",i,vdmax[i],vdmin[i]);
  printf("-------------------------------\n");
  for(i=0; i<natot; i++) 
    AntList[i].print();
  printf("-------------------------------\n");
  }  


//=======================================================================

void Obs::set(float dec1, float vhamin1[], float dt1, 
	      float vintime1[], int nconf1){
  int i;
  float tmp;

  dec=dec1;
  dt=dt1;
  nconf=nconf1;
  vhamin=new float[nconf];
  vintime=new float[nconf];
  vnpts=new int[nconf];
  printf("vintime=");
  for(i=0;i<nconf;i++){
    vhamin[i]=vhamin1[i];
    vintime[i]=vintime1[i];
    tmp=vintime[i]/dt;
    if(vintime[i]>=dt) vnpts[i]=(int) tmp;
    else vnpts[i]=1;
    printf("%f ",vintime[i]);
  }
  printf("\n");
}

void Obs::set(float dec1){
  dec=dec1;
}

void Obs::get(float *dec1, float *vhamin1, float *dt1, 
	      float *vintime1, int *nconf1){
  
  int i;
    *dec1=dec; *dt1=dt; *nconf1=nconf;
  for(i=0;i<nconf;i++){
    vhamin1[i]=vhamin[i];
    vintime1[i]=vintime[i];
  }
}

void Obs::operator=(Obs obs2){
  dec=obs2.dec; dt=obs2.dt; nconf=obs2.nconf;
  vhamin=obs2.vhamin; vintime=obs2.vintime;
  vnpts=obs2.vnpts;
}

void Obs::print(){
  int i;
  printf("\n-------------------------------\n");
  printf("        PRINT OBS\n");
  printf("-------------------------------\n");
  printf("dec= %f \n",dec);
  printf("dt=%f \n",dt);
  for(i=0;i<nconf;i++){
    printf("hamin=%7.1f itegration time=%7.4f\n",vhamin[i],vintime[i]);
  }
  printf("-------------------------------\n");
}

//=======================================================================

Shadow::Shadow(){
  pcosha=NULL;
  psinha=NULL;
}
Shadow::~Shadow(){
  if(pcosha!=NULL) {
    delete(pcosha);
    pcosha=NULL;
  }
  if(psinha!=NULL) {
    delete(psinha);
    psinha=NULL;
  }
}
void Shadow::set(float dec,float lat, float elevmin){
  float tmp,ha,hamin,dt;
  int i,isample;
  
  printf("setting Shadow\n");
  npts=1;
  hamin=0.;
  dt=0.;
  pcosha=new float[npts];
  psinha=new float[npts];
  sinlat=sin(lat*PI/180.);
  sindec=sin(dec*PI/180.);
  cosdl=cos(dec*PI/180.)*cos(lat*PI/180.); sindl=sindec*sinlat;
  /*
  tmp=(sin(elevmin*PI/180.)-sindl)/cosdl;
  if(fabs(tmp)>1){
    printf("shad: Observation always possible: no minimum for hour angle\n");
    hamin=-12*HA2RAD;
  }
  else{
    hamin=-fabs(acos(tmp));
    printf("shad :minimum possible hour angle: hamin=%f (h)\n",hamin/HA2RAD);
  }
   dt=2*hamin/(npts-1);
 */
  for(isample=0;isample<npts;isample++){
    ha=hamin+isample*dt;
    pcosha[isample]=cos(ha);
    psinha[isample]=sin(ha);
  }
}

float Shadow::getruvmin(float blx, float bly){
  int isample;
  float u,v,ruv, ruvmin;
  
  ruvmin=1.e33;
  for(isample=0;isample<npts;isample++){
    u=blx*pcosha[isample]-bly*sinlat*psinha[isample];
    v=blx*sindec*psinha[isample]+bly*(sindl*pcosha[isample]+cosdl);
    ruv=u*u+v*v;
    if(ruv<ruvmin) ruvmin=ruv;
  }
  return sqrt(ruvmin);
}


//=======================================================================

Transf::Transf(){
  pcosha=matrix(NPTSMAX,NCONFMAX);
  psinha=matrix(NPTSMAX,NCONFMAX);
  vnpts=NULL;
}
Transf::~Transf(){
  if(pcosha!=NULL) delete_matrix(pcosha);
  if(psinha!=NULL) delete_matrix(psinha);
  if(vnpts!=NULL) delete(vnpts);
  vnpts=NULL;
}
void Transf::set(Obs *pobs1, Site *psite1){
  float lat, dec, tmp, tmp2, dt, ha, elevmin, hamin;
  float vhamin[NCONFMAX],vintime[NCONFMAX];
  int nconf,i,isample;
  
  printf("setting Transf\n");
  shadowing=0;
  pobs=pobs1; psite=psite1;
  lat=psite->getlat();
  elevmin=psite->getelevmin();
  coslat=cos(lat*PI/180.);
  sinlat=sin(lat*PI/180.);
  pobs->get(&dec, vhamin, &dt, vintime, &nconf);
  if(fabs(dec-lat)>(90-2*elevmin)){
    printf("ERROR: this declination (%f) \n",dec);
    printf("       can not be observed from the site latitude (%f) !\n",lat);
    exit(1);
  }
  cosdec=cos(dec*PI/180.);
  sindec=sin(dec*PI/180.);
  cosdl=cosdec*coslat; sindl=sindec*sinlat;
  tmp2=(sin(elevmin*PI/180.)-sindl)/cosdl;
  if(fabs(tmp2)>1){
    printf("Observation always possible: hamin=-12(h)\n");
    hamin=-12*HA2RAD;
  }
  else{
    hamin=-fabs(acos(tmp2));
    printf("minimum hour angle possible for dec=%-6.1f(deg):\n",dec);
    printf("                              hamin=%-6.1f (h)\n",hamin/HA2RAD);
  }
  vnpts=new int[nconf];
  for(i=0;i<nconf;i++){
    tmp=vintime[i]/dt;
    if(vintime[i]>=dt) vnpts[i]=(int) tmp;
    else vnpts[i]=1;
    for(isample=0;isample<vnpts[i];isample++){
      ha=(vhamin[i]+isample*dt)*HA2RAD;
      if(fabs(ha)>fabs(hamin)){
	printf("ERROR: this hour angle is not possible:%-6.1f (h) \n",
	       ha/HA2RAD);
	exit(1);
      }
      pcosha[isample][i]=cos(ha);
      psinha[isample][i]=sin(ha);
    }
  }
}

void Transf::setshadow(float dec){
  float elevmin,lat;

  shadowing=1;
  lat=psite->getlat();
  elevmin=psite->getelevmin();
  shad.set(dec, lat, elevmin);
}

void Transf::xy2uv(float blx, float bly, int conf, int isample, 
		   float *pu, float *pv){
  *pu=blx*pcosha[isample][conf]-bly*sinlat*psinha[isample][conf];
  *pv=blx*sindec*psinha[isample][conf]+
    bly*(sindl*pcosha[isample][conf]+cosdl);
}
void Transf::uv2xy(float u, float v, int conf, int isample, 
		   float *px, float *py){
  float denom;
  denom = sindl+cosdl*pcosha[isample][conf];
  *px=(u*(sindl*pcosha[isample][conf]+cosdl)+
       v*sinlat*psinha[isample][conf])/denom;
  *py=(-u*sindec*psinha[isample][conf]+
       v*pcosha[isample][conf])/denom;
}
float Transf::getruvmax(float blx, float bly, int conf){
  int isample;
  float u,v,ruv, ruvmax;
  
  ruvmax=0.;
  for(isample=0;isample<vnpts[conf];isample++){
    u=blx*pcosha[isample][conf]-bly*sinlat*psinha[isample][conf];
    v=blx*sindec*psinha[isample][conf]+
      bly*(sindl*pcosha[isample][conf]+cosdl);
    ruv=u*u+v*v;
    if(ruv>ruvmax) ruvmax=ruv;
  }
  return sqrt(ruvmax);
}
float Transf::getruvmin(float blx, float bly, int conf){
  int isample;
  float u,v,ruv, ruvmin, ruvminsh;
  
  ruvmin=1.e33;
  for(isample=0;isample<vnpts[conf];isample++){
    u=blx*pcosha[isample][conf]-bly*sinlat*psinha[isample][conf];
    v=blx*sindec*psinha[isample][conf]+
      bly*(sindl*pcosha[isample][conf]+cosdl);
    ruv=u*u+v*v;
    if(ruv<ruvmin) ruvmin=ruv;
  }
  ruvmin=sqrt(ruvmin);
  if(!shadowing) return ruvmin;
  ruvminsh=shad.getruvmin(blx, bly);
  //printf("ruvminsh=%f\n",ruvminsh);
  if(ruvmin<ruvminsh) return ruvmin;
  return ruvminsh;
}

void Transf::get(float *coslat1, float *sinlat1, 
		 float *cosdec1, float *sindec1, 
		 float *cosdl1, float *sindl1){
  *coslat1=coslat; *sinlat1=sinlat;
  *cosdec1=cosdec; *sindec1=sindec;
  *cosdl1=cosdl; *sindl1=sindl;
}
//=======================================================================

void Map::set(char *filename1, float xcenter1,float ycenter1,
	      float cdelt11, float cdelt21) {

  fitsfile *fptr;  // pointer to the FITS file, defined in fitsio.h 
  int status,  nfound, anynull;
  long naxes[2], fpixel, nbuffer, npixels, ii,jj;
  float datamin, datamax, nullval;
  
  status = 0;
  strcpy(filename,filename1);
  if ( fits_open_file(&fptr, filename, READONLY, &status) )
    printfitserror( status );
  
    /* read the NAXIS1 and NAXIS2 keyword to get image size */
  if ( fits_read_keys_lng(fptr, "NAXIS", 1, 2, naxes, &nfound, &status) )
    printfitserror( status );
  printf("naxes=%ld %ld\n",naxes[0],naxes[1]);
  parray=matrix( naxes[1],naxes[0]);
  npixels  = naxes[0] * naxes[1];   // number of pixels in the image 
  fpixel   = 1;
  nullval  = 0;     // don't check for null values in the image 
  printf("reading fits file %s\n",filename);
  
  if ( fits_read_2d_flt(fptr, 0, nullval,
			naxes[0],naxes[0],naxes[1],parray[0], &anynull, &status) )
    printfitserror( status );
  
  if ( fits_close_file(fptr, &status) )
    printfitserror( status );
  naxis1=naxes[0]; naxis2=naxes[1];
  xcenter=xcenter1; ycenter=ycenter1; 
  cdelt1=cdelt11; cdelt2=cdelt21;
}

int Map::check(float x, float y){ // return 1 if forbidden

  int ix, iy;

  ix = int ((x+xcenter)/cdelt1);
  iy = int ((y+ycenter)/cdelt2);
  if(ix<0 || iy<0 || ix>=naxis1 || iy>=naxis2) return 1;
  return (int) (parray[iy][ix]+1)%2;

}

//=======================================================================

void Site::set(char *filename1, 
	       float xcenter1,float ycenter1, 
	       float cdelt11, float cdelt21,
	       float lat1, float tau1,
	       float rmin1,float rmax1){

  printf("setting site with map in %s\n",filename1);
  ground.set(filename1, xcenter1,ycenter1,cdelt11, cdelt21);
  lat=lat1; tau=tau1; nomap=0;
  rmin=rmin1; rmax=rmax1;
  elevmin=MINELEVATION;
}

int Site::check(float x, float y){ //return 1 if the position is forbidden
  int i,j;
  float r;
  r=sqrt(x*x+y*y);
  if(r>rmax || r<rmin) return 1;
  if(nomap==0) return ground.check(x,y);
  return 0;
}

//void Site::random(float *px, float *py){
void Site::print(){
  int i;
  printf("\n-------------------------------\n");
  printf("        PRINT SITE\n");
  printf("-------------------------------\n");
  printf("latitude=%7.2f \n",lat);
  printf("opacity=%6.2f \n",tau);
  if(nomap==1) printf("no map\n");
  else printf("with map\n");
  printf("rmin= %7.1f \n",rmin);
  printf("rmax= %7.1g \n",rmax);
  printf("-------------------------------\n");
  }  

//=======================================================================

void Model::set(float fwhm1, float dmin1, float dmax1){
  fwhm=fwhm1; dmin=dmin1; dmax=dmax1;
  sigsq=pow(fwhm/(2.*sqrt(2.*log(2.))),2.);
  norm=1.;
  norm=1./integral(dmin,dmax,2.*PI);
}

  
//=======================================================================


void Grid::set(int size1, float r0, float rmax, Model *mod1){
  int i;
  if(size1>=MAXDENSITYSIZE-1){
    printf("ERROR: size of grid too big\n");
    exit(1);
  }
  size=size1; mod=mod1; daz=PI/(2.*size);
  integr=mod->integral(r0,rmax,daz)/size;
  radius[0]=r0;
  az[0]=0.;  
  for(i=1;i<=size;i++) radius[i]=mod->getnextr(radius[i-1], integr, daz);
  for(i=1;i<=2*size;i++) az[i]=az[i-1]+daz;
}

void Grid::setrmax(float rmax){
  int i;

  integr=mod->integral(radius[0],rmax,daz)/size;
  az[0]=0.;  
  for(i=1;i<=size;i++) radius[i]=mod->getnextr(radius[i-1], integr, daz);
}

void Grid::getixiy(float *p, float *q, float r, float t, float u, float v){
  if(r<radius[0] || r>=radius[size]) *q=-1.;
  else *q=(mod->integral(radius[0],r,daz)/integr);
  if(v<0. || (v==0. && u<0.)) *p= -1.;
  else *p=t/daz;
}
float Grid::getdr(float iy){
    if(iy>=1) 
      return 10.*(mod->getnextr(radius[0],(iy)*integr,daz) -
	    mod->getnextr(radius[0],(iy-0.1)*integr,daz));
    return 10.*(mod->getnextr(radius[0],(iy+0.1)*integr,daz) -
	    mod->getnextr(radius[0],(iy)*integr,daz));
  };
void Grid::print(){
  int i;
  printf("\n-------------------------------\n");
  printf("        PRINT GRID\n");
  printf("-------------------------------\n");
  printf("size= %d \n",size);
  printf("integr=%f \n",integr);
  for(i=0;i<=size;i++){
    printf("rad= %7.1f ang= %7.4f\n",radius[i],az[i]);
  }
  printf("-------------------------------\n");
}

void Grid::print(char *name){
  int i;
  FILE *fp;

  if((fp=fopen(name,"w")) == NULL){
    printf("ERROR: cannot open file %s\n",name);
    exit(1);
  }
  for(i=0;i<=size;i++)
    fprintf(fp,"%-7.1f %-7.4f\n",radius[i],az[i]);
  fclose(fp);
}
//=======================================================================

void Configfile::read( int *tmp){
  char c='\0';

  fscanf(fp,"%d",tmp);
  while((c=getc(fp))!='\n' && c!=EOF);
}

void Configfile::read( long *tmp){
  char c='\0';

  fscanf(fp,"%ld",tmp);
  while((c=getc(fp))!='\n' && c!=EOF);
}

void Configfile::read( float *tmp){
  char c='\0';

  fscanf(fp,"%f",tmp);
  while((c=getc(fp))!='\n' && c!=EOF);
}

void Configfile::read( char *s){
  char c='\0';

  fscanf(fp,"%s",s);
  while((c=getc(fp))!='\n' && c!=EOF);
}

void Configfile::read(float *pv,int n){
  char c='\0';
  int i;

  for(i=0;i<n;i++) fscanf(fp,"%f",pv+i);
  while((c=getc(fp))!='\n' && c!=EOF);
}
void Configfile::read(int *pv,int n){
  char c='\0';
  int i;

  for(i=0;i<n;i++) fscanf(fp,"%d",pv+i);
  while((c=getc(fp))!='\n' && c!=EOF);
}

void Configfile::read(Instrume *pinstr, Site *psite, Obs *pobs, Model *pmod){

  char c;
  int i;

  int na,nconf,nsh,natot;
  long nvisi;
  float dec, lat, diam, tol,dt,tau, fwhm, dspec;
  float vdmax[NCONFMAX], vdmin[NCONFMAX], vhamin[NCONFMAX],vintime[NCONFMAX];
  float dmaxmod, dminmod;
  int bg, tmp, *vfixed;
  float dmask, radmin, dptmax, xmapsz, ymapsz, xmapc, ymapc;
  float *xx,*yy,*vdiam;
  char mapfile[40];


  if((fp=fopen(name,"r")) == NULL){
    printf("ERROR: cannot open file %s\n",name);
    exit(1);
  }
  printf("reading configuration in %s\n",name);

  while((c=getc(fp))!='\n');
  read(&na);  // printf("na=%d\n",na);
  read(&nconf);  // printf("nconf=%d\n",nconf);
  read(&nsh);  // printf("nsh=%d\n",nsh);
  read(&nvisi); // printf("nvisi=%d\n",nvisi);
  read(&lat); // printf("lat=%f\n",lat);
  read(&tau); // printf("tau=%f\n",tau);
  read(&dec); // printf("dec=%f\n",dec);

  read(vhamin,nconf); 
  // qsort(vhamin,nconf,sizeof(float),compf);
  // printf("vhamin="); 
  for(i=0;i<nconf;i++)printf("%f ",vhamin[i]); printf("\n");

  read(vintime,nconf);
  //qsort(vintime,nconf,sizeof(int),compi);
  printf("vintime=");
  for(i=0;i<nconf;i++)printf("%f ",vintime[i]); printf("\n");
  dt=0.0;
  read( &dt); //printf("dt=%f\n",dt);

  read(vdmax,nconf);
  //qsort(vdmax,nconf,sizeof(float),compf);
  printf("vdmax=");
  for(i=0;i<nconf;i++)printf("%f ",vdmax[i]);
  printf("\n");
  read(&tol); //printf("tol=%f\n",tol);
  read(&dmaxmod); //printf("dmaxmod=%f\n",dmaxmod);
  read(&diam); //printf("diam=%f\n",diam);
  read(&dspec);
  read(vdmin,nconf);
  //qsort(vdmin,nconf,sizeof(float),compf);
  printf("vdmin=");
  for(i=0;i<nconf;i++)printf("%f ",vdmin[i]);  printf("\n");

  read(&dminmod); //printf("dminmod=%f\n",dminmod);
  read(&dmask); //printf("dmask=%f\n",dmask);
  read(&radmin); //printf("radmin=%f\n",radmin);
  read(&dptmax); //printf("dptmax=%f\n",dptmax);
  read(mapfile); //printf("mapfile=%s\n",mapfile);
  read(&xmapsz); //printf("xmapsz=%f\n",xmapsz);
  read(&ymapsz); //printf("ymapsz=%f\n",ymapsz);
  read(&xmapc); //printf("xmapc=%f\n",xmapc);
  read(&ymapc); //printf("ymapc=%f\n",ymapc);
  read(&fwhm); //printf("fwhm=%f\n",fwhm);
  read(&bg); //printf("bg=%d\n",bg);

  natot=nconf*na-nsh*(nconf-1);
  //printf("natot=%d\n",natot);
  xx=new float[natot];
  yy=new float[natot];
  vfixed=new int[natot];
  vdiam=new float[natot];
  for(i=0;i<natot;i++) {
    vdiam[i]=diam;
    fscanf(fp,"%f%f%d",xx+i,yy+i,vfixed+i);
  }

  fclose(fp);

  //  for(i=0;i<natot;i++) printf("%7.1f %7.1f \n", xx[i],yy[i]);

  pinstr->set(xx, yy,vfixed, vdiam, na, nsh, nconf, dspec,
              vdmax, vdmin);  
  pobs->set(dec, vhamin, dt, vintime, nconf);
  if(bg==0) psite->set(lat, tau, radmin, 1.e33);
  else psite->set(mapfile,xmapc,ymapc,xmapsz,ymapsz,
		  lat, tau, radmin, 1.e33);
  pmod->set(fwhm,dminmod,dmaxmod);

  delete xx;
  delete yy;
  delete vfixed;
  delete vdiam;
}

void Configfile::write(Instrume *pinstr, Site *psite, Obs *pobs, Model *pmod){

  char c;
  int i;

  int na,nconf,nsh,natot;
  long nvisi=0l;
  float tmpf,dec, lat, diam, tol=1.,dt,tau, fwhm, dspec;
  float vdmax[NCONFMAX], vdmin[NCONFMAX], vhamin[NCONFMAX],vintime[NCONFMAX];
  float dmaxmod, dminmod;
  int bg, tmp, nomap, *vfixed;
  float dmask, radmin, radmax;
  float dptmax, xmapsz=0., ymapsz=0., xmapc=0., ymapc=0.;
  float *xx, *yy, *vdiam;
  char mapfile[40];
  time_t now;

  mapfile[0]='\0';
  if((fp=fopen(name,"w")) == NULL){
    printf("ERROR: cannot open file %s\n",name);
    exit(1);
  }
  printf("writing optimized configuration in %s\n",name);

  natot=pinstr->getnatot();

  xx=new float[natot];
  yy=new float[natot];
  vfixed=new int[natot];
  vdiam=new float[natot];

  pinstr->get(xx, yy, vfixed, vdiam, &na, &nsh, &nconf, &dspec, 
              vdmax, vdmin);  
  pobs->get(&dec, vhamin, &dt, vintime, &nconf);
  psite->get(&lat, &tau, &radmin, &radmax,&nomap);
  if(!nomap) psite->get(mapfile,&xmapc,&ymapc,&xmapsz,&ymapsz);
  pmod->get(&fwhm,&dminmod,&dmaxmod);

  diam=vdiam[0];
  dptmax=0.5*vdmax[nconf-1];
  for(i=0;i<nconf;i++) {
    tmpf=vintime[i]/dt;
    printf("tmpf=%f\n", tmpf);
    if(vintime[i]>dt) nvisi+=(na*(na-1)*((int) tmpf));
    else nvisi+=(na*(na-1));
  }

  bg = (nomap==1) ? 0 : 1;
  printf("bg=%d\n",bg);
  now=time(NULL);
  printf("%s\n",ctime(&now));

  fprintf(fp,"! %s",ctime(&now));
  fprintf(fp,"%d / nb of antennas\n",na);
  fprintf(fp,"%d / nb of conf\n",nconf);
  fprintf(fp,"%d / nb of common ant\n",nsh);
  fprintf(fp,"%ld / nb of visibilities\n",nvisi);
  fprintf(fp,"%-6.1f / site latitude\n",lat);
  fprintf(fp,"%-4.2f / opacity\n",tau);
  fprintf(fp,"%-6.1f / source declination\n",dec);

  for(i=0;i<nconf;i++) fprintf(fp,"%-4.1f ",vhamin[i]);
  fprintf(fp,"/ initial hour angle\n");

  for(i=0;i<nconf;i++)  fprintf(fp,"%-5.2f ",vintime[i]);
  fprintf(fp,"/ integration time\n");

  fprintf(fp,"%-5.2f / hour angle sample interval\n",dt);

  for(i=0;i<nconf;i++) fprintf(fp,"%-7.1f ",vdmax[i]);
  fprintf(fp,"/ maximal baseline\n");

  fprintf(fp,"%-3.1f / tolerance\n",tol);
  fprintf(fp,"%-7.1f / max baseline for model\n",dmaxmod);
  fprintf(fp,"%-4.1f / diameter\n",diam);
  fprintf(fp,"%-4.1f / minimum distance between antennas\n",dspec);
  for(i=0;i<nconf;i++)  fprintf(fp,"%-7.1f ",vdmin[i]);
  fprintf(fp,"/ minimum baseline\n");
  dmask=0.0;
  fprintf(fp,"%-7.1f / min baseline for model\n",dminmod);
  fprintf(fp,"%-7.1f / mask radius\n",dmask);
  fprintf(fp,"%-7.1f / minimum distance from map center\n",radmin);
  fprintf(fp,"%-7.1f / maximum authorized displacement\n",dptmax);
  fprintf(fp,"%s / name of map file\n",mapfile);
  fprintf(fp,"%-7.2f / x pixel size of map\n",xmapsz);
  fprintf(fp,"%-7.2f / y pixel size of map\n",ymapsz);
  fprintf(fp,"%-7.1f / x center of map\n",xmapc);
  fprintf(fp,"%-7.1f / y center of map\n",ymapc);
  fprintf(fp,"%-7.1f / fwhm of model\n",fwhm);
  fprintf(fp,"%d / map as background\n",bg);

  for(i=0;i<natot;i++) 
    fprintf(fp,"%11.3f %11.3f   %d\n", xx[i],yy[i],vfixed[i]);

  fclose(fp);

  delete xx;
  delete yy;
  delete vfixed;
  delete vdiam;

}

//=======================================================================

Padslist::Padslist(){
  npads=0;
  pos=matrix(NPADSMAX,2);
  conf=intmatrix(NPADSMAX,NCONFMAX);
}
Padslist::~Padslist(){

  delete_matrix(pos);
  delete_intmatrix(conf);
}
void Padslist::read(){

  long i;
  int j;
  FILE *fp;

  if((fp=fopen(filename,"r")) == NULL){
    printf("ERROR: cannot open file %s\n",filename);
    exit(1);
  }
  printf("reading npads in %s\n",filename);
  fscanf(fp,"%ld %d",&npads, &nconf);
  printf("npads=%ld\n",npads);

  for(i=0l;i<npads;i++) {
    fscanf(fp,"%f %f",&pos[i][0],&pos[i][1]);
    for(j=0;j<nconf;j++) fscanf(fp,"%d",&conf[i][j]);
  }
  fclose(fp);
}
void Padslist::write(){

  long i;
  int j;
  FILE *fp;

  if((fp=fopen(filename,"w")) == NULL){
    printf("ERROR: cannot open file %s\n",filename);
    exit(1);
  }

  fprintf(fp,"%ld %d \n",npads,nconf);

  for(i=0l;i<npads;i++) {
    fprintf(fp,"%-7.1f %-7.1f",pos[i][0],pos[i][1]);
    for(j=0;j<nconf;j++) fprintf(fp," %d",conf[i][j]);
    fprintf(fp,"\n");
  }
  fclose(fp);
}

void Padslist::write(char *filename2){

  long i;
  int j;
  FILE *fp;

  if((fp=fopen(filename2,"w")) == NULL){
    printf("ERROR: cannot open file %s\n",filename2);
    exit(1);
  }

  fprintf(fp,"%ld %d \n",npads,nconf);

  for(i=0l;i<npads;i++) {
    fprintf(fp,"%-7.1f %-7.1f",pos[i][0],pos[i][1]);
    for(j=0;j<nconf;j++) fprintf(fp," %d",conf[i][j]);
    fprintf(fp,"\n");
  }
  fclose(fp);
}

void Padslist::write2(char *filename2){

  long i;
  int j;
  FILE *fp;

  if((fp=fopen(filename2,"w")) == NULL){
    printf("ERROR: cannot open file %s\n",filename2);
    exit(1);
  }

  fprintf(fp,"%ld %d \n",npads,nconf);

  for(i=0l;i<npads;i++) {
    fprintf(fp,"%-5d %-7.1f %-7.1f",i+1,pos[i][0],pos[i][1]);
    for(j=0;j<nconf;j++) fprintf(fp," %d",conf[i][j]);
    fprintf(fp,"\n");
  }
  fclose(fp);
}

float Padslist::getdist2pad(float x, float y, int iconf,
			    float *xpad, float *ypad, int *ipad){
  float **dd;
  float *dd0;
  float tmpx, tmpy, dist2pad;
  int i,j,nok=0;

  *xpad=0.;
  *ypad=0.;
  dd=matrix(npads,2);
  dd0=dd[0];
  for(i=0;i<npads;i++){
    if(conf[i][iconf]!=1){
      tmpx=pos[i][0]-x;
      tmpy=pos[i][1]-y;
      dd[nok][0]=sqrt(tmpx*tmpx+tmpy*tmpy);
      dd[nok][1]=i;
      nok++;   
    }
  }
  if(nok==0){
    dd[0]=dd0;
    delete_matrix(dd);
    return -1.;
  }
  qsort(dd,nok,sizeof(float *),compdist);
  *ipad=(int) dd[0][1];
  *xpad=pos[*ipad][0];
  *ypad=pos[*ipad][1];
  dist2pad=dd[0][0];
  dd[0]=dd0;
  delete_matrix(dd);
  return dist2pad;
}

int Padslist::load(int ipad, int iconf){

  if(conf[ipad][iconf]==1) return 0;
  conf[ipad][iconf]=1;
  return 1;
}

/*
int Padslist::load(float x, float y, float dist, int iconf, 
		   float *xpad, float *ypad){

  float **dd;
  float *dd0;
  float tmp, tmpx,tmpy;
  int i,j,nok=0,loaded, toofar;

  dd=matrix(npads,3);
  dd0=dd[0];
  for(i=0;i<npads;i++){
    if(conf[i][iconf]!=1){
      tmpx=pos[i][0]-x;
      tmpy=pos[i][1]-y;
      tmp=sqrt(tmpx*tmpx+tmpy*tmpy);
      if(tmp<dist){
	dd[nok][0]=tmpx;
	dd[nok][1]=tmpy;
	dd[nok][2]=i;
	nok++;
      }
    }
  }
  qsort(dd,nok,sizeof(float *),compvec);
  for(i=0;i<nok;i++){
    j=(int) dd[i][2];
      if(conf[j][iconf]!=1){
	conf[j][iconf]=1;
	*xpad=pos[j][0];
	*ypad=pos[j][1];
	dd[0]=dd0;
	delete_matrix(dd);
	return 1;
      }
  }
  *xpad=0.;
  *ypad=0.;
  dd[0]=dd0;
  delete_matrix(dd);
  return 0;
}
*/
void Padslist::add(float x, float y, int iconf){
  if(npads==NPADSMAX){
    printf("ERROR: too many pads!\n");
    exit(1);
  }
  if(iconf>=nconf){
    printf("ERROR: wrong config index for pad: not so many config!\n");
    exit(1);
  }
  pos[npads][0]=x;
  pos[npads][1]=y;
  conf[npads][iconf]=1;
  npads++;
}

void Padslist::offset(float dx, float dy){
  
  long i;
  int j;
  
  for(i=0l;i<npads;i++) {
    pos[i][0]+=dx;
    pos[i][1]+=dy;
  }
}


void Padslist::print(){

  long i;
  int j;

  printf("\n-------------------------------\n");
  printf("        PRINT PAD POSITIONS\n");
  printf("-------------------------------\n");
  printf("npads=%ld nconf=%d\n",npads,nconf);
  for(i=0l;i<npads;i++) {
    printf("%-7.1f %-7.1f",pos[i][0],pos[i][1]);
    for(j=0;j<nconf;j++) printf("%d ",conf[i][j]);
    printf("\n");
  }
  printf("-------------------------------\n");
}
